# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
from qtpy import QtWidgets

from dioptas.model.DioptasModel import DioptasModel
from dioptas.model.util.calc import convert_units
from dioptas.widgets.MapWidget import MapWidget

from ..widgets.UtilityWidgets import get_progress_dialog, open_files_dialog
from .integration.phase.PhaseInPatternController import PhaseInPatternController
from .integration.overlay.OverlayInPatternController import OverlayInPatternController


class MapController(object):
    def __init__(self, widget: MapWidget, dioptas_model: DioptasModel):
        self.widget = widget
        self.model = dioptas_model

        self.phase_in_pattern_controller = PhaseInPatternController(
            self.widget.pattern_plot_widget, self.model
        )
        self.overlay_in_pattern_controller = OverlayInPatternController(
            self.widget.pattern_plot_widget, self.model.overlay_model
        )

        self.create_signals()

    def create_signals(self):
        self.widget.control_widget.load_btn.clicked.connect(self.load_btn_clicked)
        self.widget.control_widget.file_list.currentRowChanged.connect(
            self.file_list_row_changed
            # needs to be its own function, to always recall the model.map_model
            # this ensures, that the currently selected configuration is used
        )
        self.widget.map_plot_widget.mouse_left_clicked.connect(self.map_point_selected)
        self.widget.pattern_plot_widget.mouse_left_clicked.connect(self.pattern_clicked)
        self.widget.pattern_plot_widget.map_interactive_roi.sigRegionChanged.connect(
            self.pattern_roi_changed
        )
        self.widget.pattern_plot_widget.mouse_moved.connect(
            self.pattern_plot_mouse_moved
        )

        self.widget.map_plot_control_widget.map_dimension_cb.currentIndexChanged.connect(
            self.map_dimension_cb_changed
        )
        self.widget.map_plot_widget.mouse_moved.connect(self.map_plot_mouse_moved)
        self.widget.img_plot_widget.mouse_left_clicked.connect(
            self.img_plot_left_clicked
        )
        self.widget.img_plot_widget.mouse_moved.connect(self.img_plot_mouse_moved)

        self.model.map_model.map_changed.connect(self.update_map)
        self.model.map_model.map_changed.connect(self.update_file_list)
        self.model.clicked_tth_changed.connect(self.update_pattern_green_line)
        self.model.clicked_tth_changed.connect(self.update_image_green_line)
        self.model.clicked_tth_changed.connect(self.update_clicked_pos_label)
        self.model.clicked_azi_changed.connect(self.update_clicked_azi_label)

        self.activate_model_signals()

    def activate(self):
        self.activate_model_signals()
        self.configuration_selected()

    def activate_model_signals(self):
        self.model.img_changed.connect(self.update_image)
        self.model.pattern_changed.connect(self.update_pattern)
        self.model.configuration_selected.connect(self.configuration_selected)

    def deactivate(self):
        self.model.img_changed.disconnect(self.update_image)
        self.model.pattern_changed.disconnect(self.update_pattern)
        self.model.configuration_selected.disconnect(self.configuration_selected)

    def load_btn_clicked(self):
        filenames = open_files_dialog(
            self.widget,
            "Load image data file(s)",
            self.model.working_directories["image"],
        )
        if len(filenames) == 0:
            return

        progressDialog = get_progress_dialog(
            "Loading image data", "Abort Integration", 100, self.widget.map_pg_layout
        )

        def update_progress_dialog(value):
            progress_value = int(value / len(filenames) * 100)
            progressDialog.setValue(progress_value)
            QtWidgets.QApplication.processEvents()

        self.model.map_model.point_integrated.connect(update_progress_dialog)
        try:
            self.model.map_model.load(filenames)
            self.model.map_model.select_point(0, 0)
        except ValueError as e:
            QtWidgets.QMessageBox.critical(
                self.widget, "Error loading image data.", str(e)
            )
        finally:
            self.model.map_model.point_integrated.disconnect(progressDialog.setValue)
            progressDialog.close()

    def update_file_list(self):
        # get current items
        items = [
            self.widget.control_widget.file_list.item(i).text()
            for i in range(self.widget.control_widget.file_list.count())
        ]
        if items == self.model.map_model.get_filenames():
            return

        self.widget.control_widget.file_list.blockSignals(True)
        self.widget.control_widget.file_list.clear()
        filenames = self.model.map_model.get_filenames()
        if len(filenames) == 0:  # no files loaded
            return
        self.widget.control_widget.file_list.addItems(filenames)
        self.widget.control_widget.file_list.blockSignals(False)

    def update_map(self):
        if self.model.map_model.map is None:
            # clear image
            self.widget.map_plot_widget.plot_image(np.array([[], []]))
        else:
            self.widget.map_plot_widget.plot_image(
                np.flipud(self.model.map_model.map), auto_level=True
            )
            self.update_dimension_cb()

    def update_dimension_cb(self):
        dim_cb = self.widget.map_plot_control_widget.map_dimension_cb
        dim_cb.blockSignals(True)
        dim_cb.clear()
        possible_dimensions_str = [
            f"{x}x{y}" for x, y in self.model.map_model.possible_dimensions
        ]
        dim_cb.addItems(possible_dimensions_str)
        current_dimension_index = self.model.map_model.possible_dimensions.index(
            self.model.map_model.dimension
        )
        dim_cb.setCurrentIndex(current_dimension_index)
        dim_cb.blockSignals(False)

    def update_image(self):
        if self.model.img_model.img_data is None:
            # clear image
            self.widget.img_plot_widget.plot_image(np.array([[], []]))
        else:
            self.widget.img_plot_widget.plot_image(
                self.model.img_model.img_data, auto_level=True
            )
            self.plot_mask()

    def plot_mask(self):
        if self.model.current_configuration.use_mask:
            self.widget.img_plot_widget.activate_mask()
            self.widget.img_plot_widget.plot_mask(self.model.mask_model.get_mask())
        else:
            self.widget.img_plot_widget.deactivate_mask()

    def update_pattern(self):
        self.widget.pattern_plot_widget.plot_data(
            self.model.pattern.x, self.model.pattern.y, self.model.pattern.name
        )
        self.update_pattern_green_line(self.model.clicked_tth)

        cur_unit = self.model.integration_unit
        pattern_plot = self.widget.pattern_plot_widget.pattern_plot

        if cur_unit == "2th_deg":
            pattern_plot.setLabel("bottom", "2θ", "°")
            pattern_plot.invertX(False)
        elif cur_unit == "q_A^-1":
            pattern_plot.setLabel("bottom", "Q", "Å⁻¹")
            pattern_plot.invertX(False)
        elif cur_unit == "d_A":
            pattern_plot.setLabel("bottom", "d", "Å")
            pattern_plot.invertX(True)

    def update_pattern_green_line(self, pos):
        if self.model.integration_unit == "2th_deg":
            self.widget.pattern_plot_widget.set_pos_line(pos)
        else:
            wavelength = self.model.calibration_model.wavelength
            new_pos = convert_units(
                pos, wavelength, "2th_deg", self.model.integration_unit
            )
            self.widget.pattern_plot_widget.set_pos_line(new_pos)

    def update_image_green_line(self, pos):
        if not self.model.current_configuration.is_calibrated:
            return

        self.widget.img_plot_widget.set_circle_line(
            self.model.calibration_model.pattern_geometry.ttha, np.deg2rad(pos)
        )

    def file_list_row_changed(self, row):
        self.model.map_model.select_point_by_index(row)
        row, col = self.model.map_model.get_point_coordinates(row)
        map_shape = self.model.map_model.map.shape
        self.widget.map_plot_widget.set_mouse_click_position(
            col + 0.5, map_shape[0] - row - 0.5  # 0.5 are there to shift to center
        )

    def _get_mouse_row_col(self, x, y):
        x, y = np.floor(x), np.floor(y)
        row = self.widget.map_plot_widget.img_data.shape[0] - int(y) - 1
        col = int(x)
        return row, col

    def _row_col_in_map(self, row, col):
        map_shape = self.widget.map_plot_widget.img_data.shape
        if row < 0 or col < 0 or row >= map_shape[0] or col >= map_shape[1]:
            return False
        return True

    def map_point_selected(self, clicked_x, clicked_y):
        # skip when now map is loaded
        if self.model.map_model.map is None:
            return

        row, col = self._get_mouse_row_col(clicked_x, clicked_y)

        # skip when the mouse is outside of the map
        if not self._row_col_in_map(row, col):
            return

        self.model.map_model.select_point(row, col)
        ind = self.model.map_model.get_point_index(row, col)

        self.widget.control_widget.file_list.blockSignals(True)
        self.widget.control_widget.file_list.setCurrentRow(ind)
        self.widget.control_widget.file_list.blockSignals(False)

    def map_dimension_cb_changed(self, _):
        dimension_str = (
            self.widget.map_plot_control_widget.map_dimension_cb.currentText()
        )
        dimension = tuple([int(x) for x in dimension_str.split("x")])
        self.model.map_model.set_dimension(dimension)

    def map_plot_mouse_moved(self, x, y):
        # shows the information for a point inside of the map
        # since pyqtgraph gives the coordinates in the image coordinate system
        # we need to flip the y axis

        # skip when no image is loaded
        if self.widget.map_plot_widget.img_data is None:
            return

        row, col = self._get_mouse_row_col(x, y)

        # if the mouse is outside of the image, we don't want to show any information
        if not self._row_col_in_map(row, col):
            self.widget.map_plot_control_widget.mouse_x_label.setText(f"X: ")
            self.widget.map_plot_control_widget.mouse_y_label.setText(f"Y: ")
            self.widget.map_plot_control_widget.mouse_int_label.setText(f"I: ")
            self.widget.map_plot_control_widget.filename_label.setText(f"")
            return

        self.widget.map_plot_control_widget.mouse_x_label.setText(f"X: {col:.0f}")
        self.widget.map_plot_control_widget.mouse_y_label.setText(f"Y: {row:.0f}")
        self.widget.map_plot_control_widget.mouse_int_label.setText(
            f"I: {self.model.map_model.map[row, col]:.0f}"
        )

        point_info = self.model.map_model.get_point_info(row, col)
        if point_info.frame_index == 0:
            self.widget.map_plot_control_widget.filename_label.setText(
                f"{point_info.filename}"
            )
        else:
            self.widget.map_plot_control_widget.filename_label.setText(
                f"{point_info.filename} - Frame: {point_info.frame_index}"
            )

    def img_plot_left_clicked(self, x, y):
        if not self.model.current_configuration.is_calibrated:
            return

        calibration_model = self.model.calibration_model
        img_shape = self.model.img_model.img_data.shape

        x, y = np.array([y]), np.array([x])
        if 0.5 < x < img_shape[0] - 0.5 and 0.5 < y < img_shape[1] - 0.5:
            tth = calibration_model.get_two_theta_img(x, y)
            azi = calibration_model.get_azi_img(x, y)

            self.model.clicked_tth_changed.emit(np.rad2deg(tth))
            self.model.clicked_azi_changed.emit(np.rad2deg(azi))

    def img_plot_mouse_moved(self, x, y):
        image_x = self.widget.map_plot_control_widget.mouse_x_label
        image_y = self.widget.map_plot_control_widget.mouse_y_label
        image_int = self.widget.map_plot_control_widget.mouse_int_label

        if self.model.img_model.img_data is None:
            image_x.setText(f"X: ")
            image_y.setText(f"Y: ")
            image_int.setText(f"I: ")
            return

        img_shape = self.model.img_model.img_data.shape
        if 0 <= x < img_shape[1] and 0 <= y < img_shape[0]:
            image_x.setText(f"X: {x:.0f}")
            image_y.setText(f"Y: {y:.0f}")
            image_int.setText(f"I: {self.model.img_model.img_data[int(y), int(x)]:.0f}")
        else:
            image_x.setText(f"X: ")
            image_y.setText(f"Y: ")
            image_int.setText(f"I: ")

        if not self.model.current_configuration.is_calibrated:
            return

        x, y = y, x  # swap x and y for the calibration model
        img_tth = self.model.calibration_model.get_two_theta_img(x, y)
        img_tth = np.rad2deg(img_tth)
        img_azi = self.model.calibration_model.get_azi_img(x, y)
        img_azi = np.rad2deg(img_azi)

        tth_str, d_str, q_str, _ = self.get_position_strings(img_tth, "2th_deg")
        pos_widget = self.widget.pattern_footer_widget.mouse_unit_widget.cur_unit_widget
        pos_widget.tth_lbl.setText(tth_str)
        pos_widget.d_lbl.setText(d_str)
        pos_widget.q_lbl.setText(q_str)
        pos_widget.azi_lbl.setText(f"X: {img_azi:.3f}")

    def get_position_strings(
        self, x: float, current_unit: str | None = None
    ) -> tuple[str, str, str, str]:
        if current_unit is None:
            current_unit = self.model.integration_unit
        if self.model.calibration_model.is_calibrated:
            wavelength = self.model.calibration_model.wavelength
            if current_unit == "2th_deg":
                tth = x
                q_value = convert_units(tth, wavelength, "2th_deg", "q_A^-1")
                d_value = convert_units(tth, wavelength, "2th_deg", "d_A")
            elif current_unit == "q_A^-1":
                q_value = x
                tth = convert_units(q_value, wavelength, "q_A^-1", "2th_deg")
                d_value = convert_units(q_value, wavelength, "q_A^-1", "d_A")
            elif current_unit == "d_A":
                d_value = x
                q_value = convert_units(d_value, wavelength, "d_A", "q_A^-1")
                tth = convert_units(d_value, wavelength, "d_A", "2th_deg")
            else:
                tth = 0
                d_value = 0
                q_value = 0

            tth_str = "2θ:%9.3f" % tth
            d_str = "d:%9.3f" % d_value
            q_str = "Q:%9.3f" % q_value
        else:
            tth_str = "2θ: -"
            d_str = "d: -"
            q_str = "Q: -"
            if current_unit == "2th_deg":
                tth_str = "2θ:%9.3f" % x
            elif current_unit == "q_A^-1":
                q_str = "Q:%9.3f" % x
            elif current_unit == "d_A":
                d_str = "d:%9.3f" % x
        azi_str = "X: -"
        return tth_str, d_str, q_str, azi_str

    def pattern_plot_mouse_moved(self, x, _):
        tth_str, d_str, q_str, azi_str = self.get_position_strings(x)
        pos_widget = self.widget.pattern_footer_widget.mouse_unit_widget.cur_unit_widget
        pos_widget.tth_lbl.setText(tth_str)
        pos_widget.d_lbl.setText(d_str)
        pos_widget.q_lbl.setText(q_str)
        pos_widget.azi_lbl.setText(azi_str)

    def update_clicked_pos_label(self, pos):
        tth_str, d_str, q_str, azi_str = self.get_position_strings(pos)
        pos_widget = (
            self.widget.pattern_footer_widget.mouse_unit_widget.clicked_unit_widget
        )
        pos_widget.tth_lbl.setText(tth_str)
        pos_widget.d_lbl.setText(d_str)
        pos_widget.q_lbl.setText(q_str)
        pos_widget.azi_lbl.setText(azi_str)

    def update_clicked_azi_label(self, azi):
        pos_widget = (
            self.widget.pattern_footer_widget.mouse_unit_widget.clicked_unit_widget
        )
        pos_widget.azi_lbl.setText(f"X: {azi:.3f}")

    def pattern_clicked(self, x, _):
        self.widget.pattern_plot_widget.map_interactive_roi.setCenter(x)

    def pattern_roi_changed(self, interactive_roi):
        region = interactive_roi.getRegion()
        self.model.map_model.set_window(region)

    def configuration_selected(self):
        self.update_file_list()
        self.update_map()
        self.update_image()
        self.update_pattern()



from pyqtgraph.Qt import QtGui
import pyqtgraph.opengl as gl
import OpenGL.GL as ogl
import numpy as np


class CustomTextItem(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, X=None, Y=None, Z=None, text=None, color=(255, 255, 255, 255), size=16):
        gl.GLGraphicsItem.GLGraphicsItem.__init__(self)

        self.text = text
        self.X = X
        self.Y = Y
        self.Z = Z
        self.color = color
        self.size = size
        if self.size > 64:
            raise ValueError("Value of the variable 'size' cannot be greater than 64!")

    def setGLViewWidget(self, GLViewWidget):
        self.GLViewWidget = GLViewWidget

    def setText(self, text):
        self.text = text
        self.update()

    def setX(self, X):
        self.X = X
        self.update()

    def setY(self, Y):
        self.Y = Y
        self.update()

    def setZ(self, Z):
        self.Z = Z
        self.update()

    def paint(self):
        fontObject = QtGui.QFont()  # "Helvetica")
        fontObject.setPixelSize(self.size)

        self.GLViewWidget.qglColor(QtGui.QColor(*self.color))
        self.GLViewWidget.renderText(float(self.X), float(self.Y), float(self.Z), self.text)  # , fontObject)


class CustomAxis(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, parent, x=1, y=0, z=0, color=(0.9, 0.9, 0.9, .6)):
        gl.GLGraphicsItem.GLGraphicsItem.__init__(self)
        self.parent = parent
        self.x = x
        self.y = y
        self.z = z

        self.xLabel = CustomTextItem(x, y, z, "Hallo")
        self.xLabel.setGLViewWidget(self.parent)
        self.parent.addItem(self.xLabel)
        self.update()

    # def transform(self, x, y, z):
    #     self.resetTransform()
    #     self.transform(x, y, z)

    def setSize(self, x=None, y=None, z=None, size=None):
        """
        Set the size of the axes (in its local coordinate system; this does not affect the transform)
        Arguments can be x,y,z or size=QVector3D().
        """
        if size is not None:
            x = size.x()
            y = size.y()
            z = size.z()
        self.__size = [x, y, z]
        self.update()


class Custom3DAxis(gl.GLAxisItem):
    """Class defined to extend 'gl.GLAxisItem'."""

    def __init__(self, parent, color=(0, 0, 0, .6), axis=[True, True, True]):
        gl.GLAxisItem.__init__(self)
        self.parent = parent
        self.color = color
        self.ticks = []
        self._axisStat = axis
        self.diff = [0, 0, 0]

    def add_labels(self, x_label=None, y_label=None, z_label=None):
        """Adds axes labels."""
        x, y, z = self.size()
        dx, dy, dz = self.diff

        if x_label is not None:
            self.xLabel = CustomTextItem(X=x / 2, Y=-y / 20, Z=-z / 20, text=x_label)
            self.xLabel.setGLViewWidget(self.parent)
            self.parent.addItem(self.xLabel)
        if y_label is not None:
            self.yLabel = CustomTextItem(X=-x / 20 + dx, Y=y / 2 + dy, Z=-z / 20, text=y_label)
            self.yLabel.setGLViewWidget(self.parent)
            self.parent.addItem(self.yLabel)
        if z_label is not None:
            self.zLabel = CustomTextItem(X=-x / 20, Y=-y / 20, Z=z / 2, text=z_label)
            self.zLabel.setGLViewWidget(self.parent)
            self.parent.addItem(self.zLabel)

    def set_tick_values(self, xticks=[], yticks=[], zticks=[]):
        """Adds ticks values."""

        self.clean_ticks()
        x, y, z = self.size()
        dx, dy, dz = self.diff

        xtpos = np.linspace(0, x, len(xticks))
        ytpos = np.linspace(0, y, len(yticks))
        ztpos = np.linspace(0, z, len(zticks))
        for i, xt in enumerate(xticks):
            val = CustomTextItem(X=xtpos[i], Y=-y / 20, Z=-z / 20, text=str(xt))
            val.setGLViewWidget(self.parent)
            self.ticks.append(val)
            self.parent.addItem(val)
        for i, yt in enumerate(yticks):
            val = CustomTextItem(X=-x / 20 + dx, Y=ytpos[i], Z=-z / 20, text=str(yt))
            val.setGLViewWidget(self.parent)
            self.ticks.append(val)
            self.parent.addItem(val)
        for i, zt in enumerate(zticks):
            val = CustomTextItem(X=-x / 20, Y=-y / 20, Z=ztpos[i], text=str(zt))
            val.setGLViewWidget(self.parent)
            self.parent.addItem(val)

        self.update()

    def clean_ticks(self):
        for tick in self.ticks:
            self.parent.removeItem(tick)
        self.ticks = []

    def paint(self):
        self.setupGLState()
        if self.antialias:
            ogl.glEnable(ogl.GL_LINE_SMOOTH)
            ogl.glHint(ogl.GL_LINE_SMOOTH_HINT, ogl.GL_NICEST)
        ogl.glBegin(ogl.GL_LINES)

        x, y, z = self.size()
        paint_x, paint_y, paint_z = self._axisStat
        if paint_z:
            ogl.glColor4f(*self.color)
            ogl.glVertex3f(0, 0, 0)
            ogl.glVertex3f(0, 0, z)
        if paint_y:
            ogl.glColor4f(*self.color)
            ogl.glVertex3f(0, 0, 0)
            ogl.glVertex3f(0, y, 0)
        if paint_x:
            ogl.glColor4f(*self.color)
            ogl.glVertex3f(0, 0, 0)
            ogl.glVertex3f(x, 0, 0)
            ogl.glEnd()



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from qtpy import QtWidgets

from ...CustomWidgets import LabelAlignRight, VerticalSpacerItem, HorizontalSpacerItem
from ..CustomWidgets import BrowseFileWidget


class ImageWidget(QtWidgets.QWidget):
    def __init__(self):
        super(ImageWidget, self).__init__()

        self._create_widgets()
        self._create_layout()
        self._style_widgets()
        self._set_tooltips()

    def _create_widgets(self):
        self.file_widget = BrowseFileWidget(files='Image', checkbox_text='autoprocess')
        self.file_info_btn = QtWidgets.QPushButton('File Info')
        self.move_btn = QtWidgets.QPushButton('Position')
        self.batch_btn = QtWidgets.QPushButton('Batch view')

        self.batch_mode_widget = QtWidgets.QWidget()
        self.batch_mode_lbl = LabelAlignRight("Batch Mode:")
        self.batch_mode_integrate_rb = QtWidgets.QRadioButton("integrate")
        self.batch_mode_add_rb = QtWidgets.QRadioButton("add")
        self.batch_mode_average_rb = QtWidgets.QRadioButton("average")
        self.batch_mode_image_save_rb = QtWidgets.QRadioButton("save")

    def _create_layout(self):
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(5)

        self._layout.addWidget(self.file_widget)

        self._batch_layout = QtWidgets.QHBoxLayout()
        self._batch_layout.addWidget(self.batch_mode_lbl)
        self._batch_layout.addWidget(self.batch_mode_integrate_rb)
        self._batch_layout.addWidget(self.batch_mode_add_rb)
        self._batch_layout.addWidget(self.batch_mode_average_rb)
        self._batch_layout.addWidget(self.batch_mode_image_save_rb)
        self._batch_layout.addItem(HorizontalSpacerItem())
        self._batch_layout.addWidget(self.batch_btn)
        self.batch_mode_widget.setLayout(self._batch_layout)
        self._layout.addWidget(self.batch_mode_widget)

        self._file_info_layout = QtWidgets.QHBoxLayout()
        self._file_info_layout.addWidget(self.file_info_btn)
        self._file_info_layout.addWidget(self.move_btn)
        self._file_info_layout.addSpacerItem(HorizontalSpacerItem())

        self._layout.addLayout(self._file_info_layout)
        self._layout.addSpacerItem(VerticalSpacerItem())

        self.setLayout(self._layout)

    def _style_widgets(self):
        self._batch_layout.setContentsMargins(0, 0, 0, 0)
        self.batch_mode_integrate_rb.setChecked(True)
        self.batch_btn.setFixedHeight(25)

    def _set_tooltips(self):
        self.batch_mode_add_rb.setToolTip("Adds all images together")
        self.batch_mode_average_rb.setToolTip("Averages all images")
        self.batch_mode_integrate_rb.setToolTip("Integrates all images")
        self.batch_mode_image_save_rb.setToolTip("Saves all images")


import logging
import os
import re
import pathlib

import h5py
import numpy as np
from qtpy import QtCore
from PIL import Image
from xypattern.pattern import SmoothBrucknerBackground
from xypattern import Pattern

logger = logging.getLogger(__name__)


class BatchModel(QtCore.QObject):
    """
    Class describe a model for batch integration
    """

    def __init__(self, configuration):
        super(BatchModel, self).__init__()

        self.data = None
        self.bkg = None
        self.binning = None
        self.file_map = None
        self.files = None
        self.pos_map = None
        self.pos_map_all = None
        self.n_img = None
        self.n_img_all = None
        self.raw_available = False

        self.configuration = configuration
        self.used_mask = None
        self.used_mask_shape = None
        self.used_calibration = None

    def reset_data(self):
        self.data = None
        self.bkg = None
        self.binning = None
        self.file_map = None
        self.files = None
        self.pos_map = None
        self.pos_map_all = None
        self.n_img = None
        self.n_img_all = None
        self.used_mask = None
        self.used_mask_shape = None
        self.used_calibration = None
        self.raw_available = False

    def set_image_files(self, files):
        """
        Set internal variables with respect of given list of files.

        Open each file and count number of images inside. Position of each image in the file
        and total number of images are stored in internal variables.

        :param files: List of file names including path
        """
        if files is None:
            return
        pos_map = []
        file_map = [0]
        image_counter = 0

        self.configuration.img_model.blockSignals(True)

        for i, file in enumerate(files):
            # Assume tif file contains only one image
            if file[-4:] == ".tif":
                n_img = 1
            else:
                if not os.path.exists(file):
                    return
                self.configuration.img_model.load(file)
                n_img = self.configuration.img_model.series_max
            image_counter += n_img
            pos_map += list(zip([i] * n_img, range(n_img)))
            file_map.append(image_counter)

        self.configuration.img_model.blockSignals(False)

        self.files = np.array(files)
        self.n_img_all = image_counter
        self.raw_available = True
        self.pos_map_all = np.array(pos_map)
        self.file_map = np.array(file_map)

    def try_load_old_format(self, data_file):
        self.data = data_file["data"][()]
        self.binning = data_file["binning"][()]
        self.file_map = data_file["file_map"][()]
        self.files = data_file["files"][()].astype("U")
        self.pos_map = data_file["pos_map"][()]
        self.n_img = self.data.shape[0]
        self.n_img_all = self.data.shape[0]
        logger.info("Loading data using deprecated format")

        try:
            cal_file = str(data_file.attrs["calibration"])
            if os.path.isfile(cal_file):
                self.calibration_model.load(cal_file)
        except KeyError:
            logger.info("Calibration info is not found")

        if "mask" in data_file.attrs:
            try:
                mask_file = data_file.attrs["mask"]
                self.mask_model.load_mask(mask_file)
            except FileNotFoundError:
                logger.info("Mask file is not found")

        if "bkg" in data_file:
            self.data = data_file["bkg"][()]

    def load_proc_data(self, filename):
        """
        Load diffraction patterns and metadata from h5 file

        """
        with h5py.File(filename, "r") as data_file:
            # ToDo To be removed
            if "processed/result" not in data_file:
                self.try_load_old_format(data_file)
                return
            self.data = data_file["processed/result/data"][()]
            self.binning = data_file["processed/result/binning"][()]
            self.n_img = self.data.shape[0]
            self.n_img_all = self.data.shape[0]

            if "process" not in data_file["processed"]:
                logger.info("No matching to raw data")
                return

            self.file_map = data_file["processed/process/file_map"][()]
            self.files = data_file["processed/process/files"][()].astype("U")
            self.pos_map = data_file["processed/process/pos_map"][()]

            if isinstance(data_file["processed/process/cal_file"][()], bytes):
                self.used_calibration = str(
                    data_file["processed/process/cal_file"][()].decode("utf-8")
                )
            else:
                self.used_calibration = str(data_file["processed/process/cal_file"][()])
            if os.path.isfile(self.used_calibration):
                self.configuration.calibration_model.load(self.used_calibration)

            if "mask" in data_file["processed/process/"]:
                mask = data_file["processed/process/mask"][()]
                self.configuration.mask_model.set_dimension(mask.shape)
                self.configuration.mask_model.set_mask(mask)

            if "mask_file" in data_file["processed/process/"]:
                try:
                    self.used_mask = str(data_file["processed/process/mask_file"][()])
                    mask_data = np.array(Image.open(self.used_mask))
                    self.configuration.mask_model.set_dimension(mask_data.shape)
                    self.configuration.mask_model.load_mask(self.used_mask)
                except FileNotFoundError:
                    logger.info(f"Mask file {self.used_mask} is not found")

            if "bkg" in data_file["processed/process/"]:
                self.bkg = data_file["processed/process/bkg"][()]

    def save_proc_data(self, filename):
        """
        Save diffraction patterns to h5 file
        """
        if os.path.dirname(filename) != "":
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        with h5py.File(filename, mode="w") as f:
            f.attrs["default"] = "processed"

            nxentry = f.create_group("processed")
            nxentry.attrs["NX_class"] = "NXentry"
            nxentry.attrs["default"] = "result"

            nxdata = nxentry.create_group("result")
            nxdata.attrs["NX_class"] = "NXdata"
            nxdata.attrs["signal"] = "data"
            nxdata.attrs["axes"] = [".", "binning"]

            nxprocess = nxentry.create_group("process")
            nxprocess.attrs["NX_class"] = "NXprocess"

            if self.used_calibration is not None:
                nxprocess["cal_file"] = str(self.used_calibration)

            if self.used_mask is not None:
                nxprocess["mask_file"] = str(self.used_mask)
                nxprocess["mask_shape"] = self.used_mask_shape

            nxprocess["int_method"] = "csr"
            nxprocess["int_unit"] = "2th_deg"
            nxprocess["num_points"] = self.binning.shape[0]

            if self.bkg is not None:
                nxprocess.create_dataset("bkg", data=self.bkg)

            nxdata.create_dataset("data", data=self.data)
            tth = nxdata.create_dataset("binning", data=self.binning)
            tth.attrs["unit"] = "deg"
            tth.attrs["long_name"] = "two_theta (degrees)"

            nxprocess.create_dataset("pos_map", data=self.pos_map)
            nxprocess.create_dataset("file_map", data=self.file_map)
            nxprocess.create_dataset("files", data=self.files.astype("S"))

    def save_as_csv(self, filename):
        """
        Save diffraction patterns to 3-columns csv file
        """
        if os.path.dirname(filename) != "":
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        x = self.binning.repeat(self.n_img)
        y = (
            np.arange(self.n_img)[None, :]
            .repeat(self.binning.shape[0], axis=0)
            .flatten()
        )
        np.savetxt(
            filename,
            np.array(list(zip(x, y, self.data.T.flatten()))),
            delimiter=",",
            fmt="%f",
        )

    def integrate_raw_data(self, start, stop, step, use_all=False, callback_fn=None):
        """
        Integrate images from given file

        :param num_points: Numbers of radial bins
        :param start: Start image index from integration
        :param stop: Stop image index from integration
        :param step: Step along images to integrate
        :param use_all: Use all images. If False use only images, that were already integrated.
        :param callback_fn: callback function which is called each iteration with the current image number as parameter,
                            if it returns False the integration will be aborted.
        """
        intensity_data = []
        binning_data = []
        pos_map = []
        image_counter = 0
        current_file = ""

        if self.configuration.use_mask:
            if self.configuration.mask_model.filename != "":
                self.used_mask = self.configuration.mask_model.filename
            mask = self.configuration.mask_model.get_mask()
            self.used_mask_shape = mask.shape

        self.configuration.img_model.blockSignals(True)
        for index in range(start, stop, step):
            if use_all:
                file_index, pos = self.pos_map_all[index]
            else:
                file_index, pos = self.pos_map[index]
            if file_index != current_file:
                current_file = file_index
                self.configuration.calibration_model.img_model.load(
                    self.files[file_index]
                )

            self.configuration.img_model.load_series_img(pos + 1)
            self.configuration.mask_model.set_dimension(
                self.configuration.img_model.img_data.shape
            )

            binning, intensity = self.configuration.integrate_image_1d()
            image_counter += 1
            pos_map.append((file_index, pos))
            intensity_data.append(intensity)
            binning_data.append(binning)

            if callback_fn is not None:
                if not callback_fn(image_counter):
                    break

        self.configuration.img_model.blockSignals(False)

        # deal with different x lengths due to trimmed zeros:
        binning_lengths = [len(binning) for binning in binning_data]
        binning_max_length_ind = np.argmax(binning_lengths)
        binning_max_length = binning_lengths[binning_max_length_ind]
        binning = binning_data[binning_max_length_ind]

        for ind in range(len(intensity_data)):
            intensity_data[ind] = np.append(
                intensity_data[ind],
                np.zeros((binning_max_length - binning_lengths[ind], 1)),
            )

        # finish and save everything

        if self.configuration.calibration_model.filename != "":
            self.used_calibration = self.configuration.calibration_model.filename
        self.pos_map = np.array(pos_map)
        self.binning = np.array(binning)
        self.data = np.array(intensity_data)
        self.bkg = None
        self.n_img = self.data.shape[0]

    def extract_background(self, parameters, callback_fn=None):
        """
        Subtract background calculated with respect of given parameters
        """

        bkg = np.zeros(self.data.shape)
        auto_bkg = SmoothBrucknerBackground(*parameters)
        for i, y in enumerate(self.data):
            if callback_fn is not None:
                if not callback_fn(i):
                    break
            bkg[i] = auto_bkg.extract_background(Pattern(self.binning, y))
        self.bkg = bkg

    def normalize(self, range_ind=(10, 30)):
        if self.data is None:
            return
        average_intensities = np.mean(self.data[:, range_ind[0] : range_ind[1]], axis=1)
        factors = average_intensities[0] / average_intensities
        self.data = (self.data.T * factors).T

    def get_image_info(self, index, use_all=False):
        """
        Get filename and image position in the file

        :param index: Index of image in the batch
        :param use_all: Indexing with respect to all images. If False count only images, that were integrated.
        """
        if use_all:
            if not self.raw_available:
                return None, None
            f_index, pos = self.pos_map_all[index]
        else:
            if self.pos_map is None:
                return "NA", index
            f_index, pos = self.pos_map[index]
        filename = self.files[f_index]
        return filename, pos

    def load_image(self, index, use_all=False):
        """
        Load image in image model

        :param index: Index of image in the batch
        :param use_all: Indexing with respect to all images. If False count only images, that were integrated.
        """
        if not self.raw_available:
            return
        filename, pos = self.get_image_info(index, use_all)
        self.configuration.calibration_model.img_model.load(filename, pos)

    def get_next_folder_filenames(self):
        """
        Loads all files from the next folder with similar file-endings.
        """
        folder_path, _ = os.path.split(self.files[0])
        next_folder_path = iterate_folder(folder_path, 1)
        files = []
        if next_folder_path is not None and os.path.exists(next_folder_path):
            for file in os.listdir(next_folder_path):
                if file.endswith(pathlib.Path(self.files[0]).suffix):
                    files.append(os.path.join(next_folder_path, file))
        files = sorted(files)
        return files[: self.n_img_all]

    def get_previous_folder_filenames(self):
        """
        Loads all files from the previous folder with similar file-endings.
        """
        folder_path, _ = os.path.split(self.files[0])
        previous_folder_path = iterate_folder(folder_path, -1)
        files = []
        if previous_folder_path is not None and os.path.exists(previous_folder_path):
            for file in os.listdir(previous_folder_path):
                if file.endswith(pathlib.Path(self.files[0]).suffix):
                    files.append(os.path.join(previous_folder_path, file))
        files = sorted(files)
        return files[: self.n_img_all]


def iterate_folder(folder_path, step):
    pattern = re.compile(r"\d+")
    match_iterator = pattern.finditer(folder_path)
    new_directory_str = None
    for ind, match in enumerate(list(match_iterator)):
        number_span = match.span()
        left_ind = number_span[0]
        right_ind = number_span[1]
        number = int(folder_path[left_ind:right_ind]) + step
        if number < 0:
            number = 0
        new_directory_str = "{left_str}{number:0{len}}{right_str}".format(
            left_str=folder_path[:left_ind],
            number=number,
            len=right_ind - left_ind,
            right_str=folder_path[right_ind:],
        )
    return new_directory_str



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .CalibrationModel import CalibrationModel
from .ImgModel import ImgModel
from .MaskModel import MaskModel
from .PhaseModel import PhaseModel
from .PatternModel import PatternModel
from .OverlayModel import OverlayModel
from .MapModel2 import MapModel2
from .BatchModel import BatchModel



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pyqtgraph.graphicsItems.GraphicsWidget import GraphicsWidget
from pyqtgraph.graphicsItems.LabelItem import LabelItem
from qtpy import QtWidgets, QtCore, QtGui
from pyqtgraph import functions as fn
from pyqtgraph.Point import Point
from pyqtgraph.graphicsItems.ScatterPlotItem import ScatterPlotItem, drawSymbol
from pyqtgraph.graphicsItems.PlotDataItem import PlotDataItem
from pyqtgraph.graphicsItems.GraphicsWidgetAnchor import GraphicsWidgetAnchor

__all__ = ["LegendItem"]


class LegendItem(GraphicsWidget, GraphicsWidgetAnchor):
    """
    Displays a legend used for describing the contents of a plot.
    LegendItems are most commonly created by calling PlotItem.addLegend().

    Note that this item should not be added directly to a PlotItem. Instead,
    Make it a direct descendant of the PlotItem::

        legend.setParentItem(plotItem)

    Codebase was copied from original pyqtgraph legenditem code. It had to be included here,
    because the Pull request for additional features needed has not been worked through yet.

    """

    def __init__(
        self,
        size=None,
        offset=None,
        horSpacing=25,
        verSpacing=0,
        box=True,
        labelAlignment="center",
        showLines=True,
    ):
        """
        ==============  ===============================================================
        **Arguments:**
        size            Specifies the fixed size (width, height) of the legend. If
                        this argument is omitted, the legend will autimatically resize
                        to fit its contents.
        offset          Specifies the offset position relative to the legend's parent.
                        Positive values offset from the left or top; negative values
                        offset from the right or bottom. If offset is None, the
                        legend must be anchored manually by calling anchor() or
                        positioned by calling setPos().
        horSpacing      Specifies the spacing between the line symbol and the label.
        verSpacing      Specifies the spacing between individual entries of the legend
                        vertically. (Can also be negative to have them really close)
        box             Specifies if the Legend should will be drawn with a rectangle
                        around it.
        labelAlignment  Specifies the alignment of the label texts. Possible values are
                        "center", "left" or "right".
        showLines       Specifies whether or not the lines should be shown in the legend.
                        If value is "False" it will only show the labels with the corresponding
                        text color.
        ==============  ===============================================================

        """

        GraphicsWidget.__init__(self)
        GraphicsWidgetAnchor.__init__(self)
        self.setFlag(self.ItemIgnoresTransformations)
        self.layout = QtWidgets.QGraphicsGridLayout()
        self.layout.setVerticalSpacing(verSpacing)
        self.layout.setHorizontalSpacing(horSpacing)
        self._horSpacing = horSpacing
        self._verSpacing = verSpacing
        self.setLayout(self.layout)
        self.legendItems = []
        self.plotItems = []
        self.hiddenFlag = []
        self.size = size
        self.offset = offset
        self.box = box
        self.label_alignment = labelAlignment
        self.showLines = showLines
        # A numItems variable needs to be introduced, because chaining removeItem and addItem function in random order,
        # will otherwise lead to writing in the same layout row. Idea here is to always insert LabelItems on larger
        # and larger layout row numbers. The GraphicsGridlayout item will not care about empty rows.
        self.numItems = 0
        if size is not None:
            self.setGeometry(QtCore.QRectF(0, 0, self.size[0], self.size[1]))

    def setParentItem(self, p):
        ret = GraphicsWidget.setParentItem(self, p)
        if self.offset is not None:
            offset = Point(self.offset)
            anchorx = 1 if offset[0] <= 0 else 0
            anchory = 1 if offset[1] <= 0 else 0
            anchor = (anchorx, anchory)
            self.anchor(itemPos=anchor, parentPos=anchor, offset=offset)
        return ret

    def addItem(self, item, name):
        """
        Add a new entry to the legend.

        ==============  ========================================================
        **Arguments:**
        item            A PlotDataItem from which the line and point style
                        of the item will be determined or an instance of
                        ItemSample (or a subclass), allowing the item display
                        to be customized.
        title           The title to display for this item. Simple HTML allowed.
        ==============  ========================================================
        """

        # get item color
        pen = fn.mkPen(item.opts["pen"])
        color = pen.color()
        color_str = color.name()

        # create label with same color
        label = LabelItem()
        label.setAttr("color", color_str)
        label.setAttr("justify", self.label_alignment)
        label.setText(name)

        if isinstance(item, ItemSample):
            sample = item
        else:
            sample = ItemSample(item)

        self.legendItems.append((sample, label))
        self.hiddenFlag.append(False)
        self.plotItems.append(item)
        if self.showLines:
            self.layout.addItem(sample, self.numItems, 0)
        self.layout.addItem(label, self.numItems, 1)
        self.numItems += 1
        self.updateSize()

    def removeItem(self, name):
        """
        Removes one item from the legend.

        ==============  ========================================================
        **Arguments:**
        name            Either the name displayed for this item or the originally
                        added item object.
        ==============  ========================================================
        """
        # Thanks, Ulrich!
        # cycle for a match
        ind = 0
        for sample, label in self.legendItems:
            if label.text == name:  # hit
                self.legendItems.remove((sample, label))  # remove from itemlist
                if not self.hiddenFlag[ind]:
                    if self.showLines:
                        self.layout.removeItem(sample)  # remove from layout
                    self.layout.removeItem(label)
                    sample.close()  # remove from drawing
                    label.close()
                self.updateSize()  # redraq box
                self.numItems -= 1
                del self.hiddenFlag[ind]
                return
            ind += 1

        for ind, item in enumerate(self.plotItems):
            if item == name:
                sample, label = self.legendItems[ind]
                self.plotItems.remove(item)

                if not self.hiddenFlag[ind]:
                    if self.showLines:
                        self.layout.removeItem(sample)
                    self.layout.removeItem(label)
                sample.close()
                label.close()
                self.legendItems.remove((sample, label))
                self.updateSize()
                self.numItems -= 1
                del self.hiddenFlag[ind]

    def hideItem(self, ind):
        sample_item, label_item = self.legendItems[ind]
        if not self.hiddenFlag[ind]:
            if self.showLines:
                self.layout.removeItem(sample_item)
                sample_item.hide()
            self.layout.removeItem(label_item)
            label_item.hide()
        self.hiddenFlag[ind] = True
        self.updateSize()

    def showItem(self, ind):
        sample_item, label_item = self.legendItems[ind]
        if self.hiddenFlag[ind]:
            if self.showLines:
                self.layout.addItem(sample_item, ind, 0)
                sample_item.show()
            self.layout.addItem(label_item, ind, 1)
            label_item.show()
        self.hiddenFlag[ind] = False
        self.updateSize()

    def setItemColor(self, ind, color):
        sample_item, label_item = self.legendItems[ind]
        label_item.setAttr("color", color)
        label_item.setText(label_item.text)

    def renameItem(self, ind, name):
        sample_item, label_item = self.legendItems[ind]
        label_item.setText(name)
        self.updateSize()

    def updateSize(self):
        if self.size is not None:
            return
        # we only need to set geometry to 0, as now the horizontal and vertical spacing is set in
        # __init__.
        self.setGeometry(0, 0, 0, 0)

    def boundingRect(self):
        return QtCore.QRectF(0, 0, self.width(), self.height())

    def paint(self, p, *args):
        if self.box:
            p.setPen(fn.mkPen(255, 255, 255, 100))
            p.setBrush(fn.mkBrush(100, 100, 100, 50))
            p.drawRect(self.boundingRect())

    def hoverEvent(self, ev):
        ev.acceptDrags(QtCore.Qt.LeftButton)

    def mouseDragEvent(self, ev):
        if ev.button() == QtCore.Qt.LeftButton:
            dpos = ev.pos() - ev.lastPos()
            self.autoAnchor(self.pos() + dpos)


class ItemSample(GraphicsWidget):
    """Class responsible for drawing a single item in a LegendItem (sans label).

    This may be subclassed to draw custom graphics in a Legend.
    """

    ## Todo: make this more generic; let each item decide how it should be represented.
    def __init__(self, item):
        GraphicsWidget.__init__(self)
        self.item = item

    def boundingRect(self):
        return QtCore.QRectF(0, 0, 20, 20)

    def paint(self, p, *args):
        # p.setRenderHint(p.Antialiasing)  # only if the data is antialiased.
        opts = self.item.opts

        if (
            opts.get("fillLevel", None) is not None
            and opts.get("fillBrush", None) is not None
        ):
            p.setBrush(fn.mkBrush(opts["fillBrush"]))
            p.setPen(fn.mkPen(None))
            p.drawPolygon(
                QtGui.QPolygonF(
                    [
                        QtCore.QPointF(2, 18),
                        QtCore.QPointF(18, 2),
                        QtCore.QPointF(18, 18),
                    ]
                )
            )

        if not isinstance(self.item, ScatterPlotItem):
            p.setPen(fn.mkPen(opts["pen"]))
            p.drawLine(2, 18, 18, 2)

        symbol = opts.get("symbol", None)
        if symbol is not None:
            if isinstance(self.item, PlotDataItem):
                opts = self.item.scatter.opts

            pen = fn.mkPen(opts["pen"])
            brush = fn.mkBrush(opts["brush"])
            size = opts["size"]

            p.translate(10, 10)
            path = drawSymbol(p, symbol, size, pen, brush)




# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from qtpy import QtWidgets, QtCore

import numpy as np

from ..widgets.UtilityWidgets import open_file_dialog, save_file_dialog
from .. import calibrants_path

# imports for type hinting in PyCharm -- DO NOT DELETE
from ..widgets.CalibrationWidget import CalibrationWidget
from ..widgets.UtilityWidgets import open_file_dialog
from ..model.DioptasModel import DioptasModel
from ..model.CalibrationModel import (
    NotEnoughSpacingsInCalibrant,
    get_available_detectors,
    DetectorModes,
)


class CalibrationController(object):
    """
    CalibrationController handles all the interaction between the CalibrationView and the CalibrationData class
    """

    def __init__(self, widget, dioptas_model):
        """Manages the connection between the calibration GUI and data

        :param widget: Gives the Calibration Widget
        :type widget: CalibrationWidget

        :param dioptas_model: Reference to DioptasModel Object
        :type dioptas_model: DioptasModel

        """
        self.widget = widget
        self.model = dioptas_model

        self.widget.set_start_values(self.model.calibration_model.start_values)
        self.create_signals()
        self.load_detectors_list()
        self.load_calibrants_list()

    def create_signals(self):
        """
        Connects the GUI signals to the appropriate Controller methods.
        """
        self.model.img_changed.connect(self.plot_image)
        self.model.configuration_selected.connect(
            self.update_calibration_parameter_in_view
        )
        self.model.configuration_selected.connect(
            self.update_detector_parameters_in_view
        )
        self.model.calibration_model.detector_reset.connect(
            self.update_detector_parameters_in_view
        )
        self.model.calibration_model.detector_reset.connect(
            self.show_detector_reset_message_box
        )

        self.create_transformation_signals()
        self.create_update_signals()
        self.create_mouse_signals()

        self.widget.detectors_cb.currentIndexChanged.connect(self.load_detector)
        self.widget.detector_load_btn.clicked.connect(self.load_detector_from_file)
        self.widget.detector_reset_btn.clicked.connect(self.reset_detector_from_file)

        self.widget.calibrant_cb.currentIndexChanged.connect(self.load_calibrant)
        self.widget.load_img_btn.clicked.connect(self.load_img)
        self.widget.load_next_img_btn.clicked.connect(self.load_next_img)
        self.widget.load_previous_img_btn.clicked.connect(self.load_previous_img)
        self.widget.filename_txt.editingFinished.connect(self.update_filename_txt)

        self.widget.save_calibration_btn.clicked.connect(self.save_calibration)
        self.widget.load_calibration_btn.clicked.connect(self.load_calibration)
        self.widget.calibrate_btn.clicked.connect(self.calibrate)
        self.widget.refine_btn.clicked.connect(self.refine)

        self.widget.clear_peaks_btn.clicked.connect(self.clear_peaks)
        self.widget.undo_peaks_btn.clicked.connect(self.undo_peaks_btn_clicked)

        self.widget.load_spline_btn.clicked.connect(self.load_spline_btn_click)
        self.widget.spline_reset_btn.clicked.connect(self.reset_spline_btn_click)

        self.widget.f2_wavelength_cb.stateChanged.connect(self.wavelength_cb_changed)
        self.widget.pf_wavelength_cb.stateChanged.connect(self.wavelength_cb_changed)
        self.widget.sv_wavelength_cb.stateChanged.connect(self.wavelength_cb_changed)

        self.widget.f2_distance_cb.stateChanged.connect(self.distance_cb_changed)
        self.widget.pf_distance_cb.stateChanged.connect(self.distance_cb_changed)
        self.widget.sv_distance_cb.stateChanged.connect(self.distance_cb_changed)

        self.widget.use_mask_cb.stateChanged.connect(self.plot_mask)
        self.widget.mask_transparent_cb.stateChanged.connect(
            self.mask_transparent_status_changed
        )

    def create_transformation_signals(self):
        """
        Connects all the rotation GUI controls.
        """
        self.widget.rotate_m90_btn.clicked.connect(self.rotate_m90_btn_clicked)
        self.widget.rotate_p90_btn.clicked.connect(self.rotate_p90_btn_clicked)
        self.widget.invert_horizontal_btn.clicked.connect(
            self.invert_horizontal_btn_clicked
        )
        self.widget.invert_vertical_btn.clicked.connect(
            self.invert_vertical_btn_clicked
        )
        self.widget.reset_transformations_btn.clicked.connect(
            self.reset_transformations_btn_clicked
        )

    def activate(self):
        if not self.model.img_changed.has_listener(self.plot_image):
            self.model.img_changed.connect(self.plot_image)
        self.update_img_plot()

    def update_img_plot(self):
        self.plot_image()
        self.plot_mask()

    def deactivate(self):
        if self.model.img_changed.has_listener(self.plot_image):
            self.model.img_changed.disconnect(self.plot_image)

    def rotate_m90_btn_clicked(self):
        self.model.calibration_model.rotate_detector_m90()
        self.model.img_model.rotate_img_m90()
        self.clear_peaks()

    def rotate_p90_btn_clicked(self):
        self.model.calibration_model.rotate_detector_p90()
        self.model.img_model.rotate_img_p90()
        self.clear_peaks()

    def invert_horizontal_btn_clicked(self):
        self.model.calibration_model.flip_detector_horizontally()
        self.model.img_model.flip_img_horizontally()
        self.clear_peaks()

    def invert_vertical_btn_clicked(self):
        self.model.calibration_model.flip_detector_vertically()
        self.model.img_model.flip_img_vertically()
        self.clear_peaks()

    def reset_transformations_btn_clicked(self):
        self.model.calibration_model.reset_transformations()
        self.model.img_model.reset_transformations()
        self.clear_peaks()

    def create_update_signals(self):
        """
        Connects all the txt box signals. Which specifically are the update buttons here.
        """
        self.widget.f2_update_btn.clicked.connect(self.update_f2_btn_click)
        self.widget.pf_update_btn.clicked.connect(self.update_pyFAI_btn_click)

    def create_mouse_signals(self):
        """
        Creates the mouse_move connections to show the current position of the mouse pointer.
        """
        self.widget.img_widget.mouse_moved.connect(self.update_img_mouse_position_lbl)
        self.widget.cake_widget.mouse_moved.connect(self.update_cake_mouse_position_lbl)
        self.widget.pattern_widget.mouse_moved.connect(
            self.update_pattern_mouse_position_lbl
        )
        self.widget.img_widget.mouse_left_clicked.connect(self.search_peaks)

    def update_f2_btn_click(self):
        """
        Takes all parameters inserted into the fit2d txt-fields and updates the current calibration accordingly.
        """
        fit2d_parameter = self.widget.get_fit2d_parameter()
        self.model.calibration_model.set_fit2d(fit2d_parameter)
        self.update_all()

    def update_pyFAI_btn_click(self):
        """
        Takes all parameters inserted into the fit2d txt-fields and updates the current calibration accordingly.
        """
        pyFAI_parameter = self.widget.get_pyFAI_parameter()
        self.model.calibration_model.set_pyFAI(pyFAI_parameter)
        self.update_all()

    def load_img(self):
        """
        Loads an image file.
        """
        filename = open_file_dialog(
            self.widget,
            caption="Load Calibration Image",
            directory=self.model.working_directories["image"],
        )

        if filename != "":
            self.model.working_directories["image"] = os.path.dirname(filename)
            self.model.img_model.load(filename)

    def load_next_img(self):
        self.model.img_model.load_next_file()

    def load_previous_img(self):
        self.model.img_model.load_previous_file()

    def update_filename_txt(self):
        """
        Updates the filename in the GUI corresponding to the filename in img_data
        """
        current_filename = os.path.basename(self.model.img_model.filename)
        current_directory = os.path.dirname(self.model.img_model.filename)
        new_filename = str(self.widget.filename_txt.text())
        if current_filename == new_filename:
            return
        if os.path.exists(os.path.join(current_directory, new_filename)):
            try:
                self.load_img(os.path.join(current_directory, new_filename))
            except TypeError:
                self.widget.filename_txt.setText(current_filename)
        else:
            self.widget.filename_txt.setText(current_filename)

    def load_detectors_list(self):
        self._detectors_list, _ = get_available_detectors()
        self._detectors_list.insert(0, "Custom")
        self.widget.detectors_cb.blockSignals(True)
        self.widget.detectors_cb.clear()
        self.widget.detectors_cb.addItems(self._detectors_list)
        self.widget.detectors_cb.insertSeparator(1)
        self.widget.detectors_cb.insertSeparator(1)
        self.widget.detectors_cb.blockSignals(False)

    def load_detector(self, ind):
        """
        Loads the selected Detector from the Detector combobox into the calibration model. This blackout disable the
        controls for pixel widths, unless "custom" (the first element) is selected.
        """
        if ind != 0:
            self.model.calibration_model.load_detector(
                self.widget.detectors_cb.currentText()
            )
            emit_img_changed = (
                self.model.calibration_model.detector.shape
                == self.model.img_model.img_data.shape
            )
            # makes no sense to have transformations when loading a detector, however only emitting that the img changed
            # if detector and image have same size, otherwise the user should have the possibility to load an image
            # without error
            self.model.img_model.reset_transformations(emit_img_changed)
        else:
            self.model.calibration_model.reset_detector()
        self.update_detector_parameters_in_view()

    def load_detector_from_file(self):
        filename = open_file_dialog(
            self.widget,
            caption="Load Nexus Detector",
            directory=self.model.working_directories["image"],
            filter="*.h5",
        )

        if filename != "":
            self.model.calibration_model.load_detector_from_file(filename)
            self.update_detector_parameters_in_view()

    def reset_detector_from_file(self):
        self.model.calibration_model.reset_detector()
        self.model.img_model.reset_transformations()
        self.update_detector_parameters_in_view()

    def _update_pixel_size_in_gui(self):
        self.widget.set_pixel_size(
            self.model.calibration_model.orig_pixel2,
            self.model.calibration_model.orig_pixel1,
        )

    def _update_spline_in_gui(self):
        if self.model.calibration_model.detector.splineFile is not None:
            self.widget.spline_filename_txt.setText(
                os.path.basename(self.model.calibration_model.detector.splineFile)
            )
        elif not self.model.calibration_model.detector.uniform_pixel:
            self.widget.spline_filename_txt.setText("from Detector")
        else:
            self.widget.spline_filename_txt.setText("None")

    def load_calibrants_list(self):
        """
        Loads all calibrants from the ExampleData/calibrants directory into the calibrants combobox. And loads number 7.
        """
        self._calibrants_file_list = []
        self._calibrants_file_names_list = []
        for file in os.listdir(calibrants_path):
            if file.endswith(".D"):
                self._calibrants_file_list.append(file)
                self._calibrants_file_names_list.append(file.split(".")[:-1][0])
        self._calibrants_file_list.sort()
        self._calibrants_file_names_list.sort()
        self.widget.calibrant_cb.blockSignals(True)
        self.widget.calibrant_cb.clear()
        self.widget.calibrant_cb.addItems(self._calibrants_file_names_list)
        self.widget.calibrant_cb.blockSignals(False)
        self.widget.calibrant_cb.setCurrentIndex(
            self._calibrants_file_names_list.index("LaB6")
        )  # to LaB6
        self.load_calibrant()

    def load_calibrant(self, wavelength_from="start_values"):
        """
        Loads the selected calibrant in the calibrant combobox into the calibration data.
        :param wavelength_from: determines which wavelength to use possible values: "start_values", "pyFAI"
        """
        current_index = self.widget.calibrant_cb.currentIndex()
        filename = os.path.join(
            self.model.calibration_model._calibrants_working_dir,
            self._calibrants_file_list[current_index],
        )
        self.model.calibration_model.set_calibrant(filename)

        if wavelength_from == "start_values":
            start_values = self.widget.get_start_values()
            wavelength = start_values["wavelength"]
        elif wavelength_from == "pyFAI":
            pyFAI_parameter, _ = (
                self.model.calibration_model.get_calibration_parameter()
            )
            if pyFAI_parameter["wavelength"] != 0:
                wavelength = pyFAI_parameter["wavelength"]
            else:
                start_values = self.widget.get_start_values()
                wavelength = start_values["wavelength"]
        else:
            start_values = self.widget.get_start_values()
            wavelength = start_values["wavelength"]

        self.model.calibration_model.calibrant.setWavelength_change2th(wavelength)
        try:
            integration_unit = self.model.current_configuration.integration_unit
        except:
            integration_unit = "2th_deg"

        calibrant_line_positions = self.convert_x_value(
            np.array(self.model.calibration_model.calibrant.get_2th()) / np.pi * 180,
            "2th_deg",
            integration_unit,
            wavelength,
        )
        # filter them to only show the ones visible with the current pattern
        if len(self.model.pattern.x) > 0:
            pattern_min = np.min(self.model.pattern.x)
            pattern_max = np.max(self.model.pattern.x)
            calibrant_line_positions = calibrant_line_positions[
                calibrant_line_positions > pattern_min
            ]
            calibrant_line_positions = calibrant_line_positions[
                calibrant_line_positions < pattern_max
            ]
            self.widget.pattern_widget.plot_vertical_lines(
                positions=calibrant_line_positions,
                name=self._calibrants_file_names_list[current_index],
            )

    def set_calibrant(self, index):
        """
        :param index:
            index of a specific calibrant in the calibrant combobox
        """
        self.widget.calibrant_cb.setCurrentIndex(index)
        self.load_calibrant()

    def plot_image(self, autoscale=True):
        """
        Plots the current image loaded in img_data and autoscales the intensity.
        """
        self.widget.img_widget.plot_image(self.model.img_data, autoscale)
        self.widget.set_img_filename(self.model.img_model.filename)

    def search_peaks(self, x, y):
        """
        Searches peaks around a specific points (x,y) in the current image file. The algorithm for searching
        (either automatic or single peaksearch) is set in the GUI.
        :param x:
            x-Position for the search.
        :param y:
            y-Position for the search
        """
        x, y = (
            y,
            x,
        )  # indices for the img array are transposed compared to the mouse position

        # convert pixel coord into pixel index
        x, y = int(x), int(y)

        # filter events outside the image
        shape = self.model.img_model.img_data.shape
        if not (0 <= x < shape[0]):
            return
        if not (0 <= y < shape[1]):
            return

        peak_ind = self.widget.peak_num_sb.value()
        if self.widget.automatic_peak_search_rb.isChecked():
            points = self.model.calibration_model.find_peaks_automatic(
                x, y, peak_ind - 1
            )
        else:
            search_size = int(self.widget.search_size_sb.value())
            points = self.model.calibration_model.find_peak(
                x, y, search_size, peak_ind - 1
            )
        if len(points):
            self.plot_points(points)
            if self.widget.automatic_peak_num_inc_cb.isChecked():
                self.widget.peak_num_sb.setValue(peak_ind + 1)

    def plot_points(self, points=None):
        """
        Plots points into the image view.
        :param points:
            list of points, whereby a point is a [x,y] element. If it is none it will plot the points stored in the
            calibration_data
        """
        if points is None:
            try:
                points = self.model.calibration_model.get_point_array()
            except IndexError:
                points = []
        if len(points):
            self.widget.img_widget.add_scatter_data(
                points[:, 0] + 0.5, points[:, 1] + 0.5
            )

    def clear_peaks(self):
        """
        Deletes all points/peaks in the calibration_data and in the GUI.
        """
        self.model.calibration_model.clear_peaks()
        self.widget.img_widget.clear_scatter_plot()
        self.widget.peak_num_sb.setValue(1)

    def undo_peaks_btn_clicked(self):
        """
        undoes clicked peaks
        """
        num_points = self.model.calibration_model.remove_last_peak()
        self.widget.img_widget.remove_last_scatter_points(num_points)
        if self.widget.automatic_peak_num_inc_cb.isChecked():
            self.widget.peak_num_sb.setValue(self.widget.peak_num_sb.value() - 1)

    def load_spline_btn_click(self):
        filename = open_file_dialog(
            self.widget,
            caption="Load Distortion Spline File",
            directory=self.model.working_directories["image"],
            filter="*.spline",
        )

        if filename != "":
            self.model.calibration_model.load_distortion(filename)
            self._update_spline_in_gui()
            self.widget.spline_reset_btn.setEnabled(True)

    def reset_spline_btn_click(self):
        self.model.calibration_model.reset_distortion_correction()
        self.widget.spline_filename_txt.setText("None")
        self.widget.spline_reset_btn.setEnabled(False)

    def wavelength_cb_changed(self, value):
        """
        Sets the fit_wavelength parameter in the calibration data according to the GUI state.
        """
        self.widget.f2_wavelength_cb.blockSignals(True)
        self.widget.pf_wavelength_cb.blockSignals(True)
        self.widget.sv_wavelength_cb.blockSignals(True)

        self.widget.f2_wavelength_cb.setChecked(value)
        self.widget.pf_wavelength_cb.setChecked(value)
        self.widget.sv_wavelength_cb.setChecked(value)

        self.widget.f2_wavelength_cb.blockSignals(False)
        self.widget.pf_wavelength_cb.blockSignals(False)
        self.widget.sv_wavelength_cb.blockSignals(False)

        self.model.calibration_model.fit_wavelength = value

    def distance_cb_changed(self, value):
        """
        Sets the fit_distance parameter int he calibration model according to the GUI state
        """
        self.widget.f2_distance_cb.blockSignals(True)
        self.widget.pf_distance_cb.blockSignals(True)
        self.widget.sv_distance_cb.blockSignals(True)

        self.widget.f2_distance_cb.setChecked(value)
        self.widget.pf_distance_cb.setChecked(value)
        self.widget.sv_distance_cb.setChecked(value)

        self.widget.f2_distance_cb.blockSignals(False)
        self.widget.pf_distance_cb.blockSignals(False)
        self.widget.sv_distance_cb.blockSignals(False)

        self.model.calibration_model.fit_distance = value

    def update_fixed_values(self):
        self.model.calibration_model.set_fixed_values(self.widget.get_fixed_values())

    def calibrate(self):
        """
        Performs calibration based on the previously inputted/searched peaks and start values.
        """
        if len(self.model.calibration_model.points) == 0:
            QtWidgets.QMessageBox.critical(
                self.widget,
                "No peaks defined!",
                "No peaks defined. Please define initial peaks first as a starting point.<br><br><a href='https://dioptas.readthedocs.io/en/stable/calibration.html'>Consult the manual for more information. </a>",
                QtWidgets.QMessageBox.Ok,
            )
            return
        self.load_calibrant()  # load the right calibration file...
        self.model.calibration_model.set_start_values(self.widget.get_start_values())
        self.model.calibration_model.set_pixel_size(self.widget.get_pixel_size())
        self.model.calibration_model.set_fixed_values(self.widget.get_fixed_values())
        progress_dialog = self.create_progress_dialog(
            "Calibrating.", "", 0, show_cancel_btn=False
        )
        self.model.calibration_model.calibrate()

        progress_dialog.close()

        if self.widget.options_automatic_refinement_cb.isChecked():
            self.automatic_refinement()
        else:
            self.update_all()
        self.update_calibration_parameter_in_view()

    def refine(self):
        self.model.calibration_model.set_fixed_values(self.widget.get_fixed_values())

        if self.widget.options_automatic_refinement_cb.isChecked():
            self.automatic_refinement()
        else:
            progress_dialog = self.create_progress_dialog(
                "Refining.", "", 0, show_cancel_btn=False
            )
            self.model.calibration_model.refine()
            progress_dialog.close()
            self.update_all()

        self.update_calibration_parameter_in_view()

    def create_progress_dialog(
        self, text_str, abort_str, end_value, show_cancel_btn=True
    ):
        """Creates a Progress Bar Dialog.
        :param text_str:  Main message string
        :param abort_str:  Text on the abort button
        :param end_value:  Number of steps for which the progressbar is being used
        :param show_cancel_btn: Whether the cancel button should be shown.
        :return: ProgressDialog reference which is already shown in the interface
        :rtype: QtWidgets.ProgressDialog
        """
        progress_dialog = QtWidgets.QProgressDialog(
            text_str, abort_str, 0, end_value, self.widget
        )

        progress_dialog.move(
            int(
                self.widget.tab_widget.x()
                + self.widget.tab_widget.size().width() / 2.0
                - progress_dialog.size().width() / 2.0
            ),
            int(
                self.widget.tab_widget.y()
                + self.widget.tab_widget.size().height() / 2.0
                - progress_dialog.size().height() / 2.0
            ),
        )

        progress_dialog.setWindowTitle("   ")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
        progress_dialog.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        if not show_cancel_btn:
            progress_dialog.setCancelButton(None)
        progress_dialog.show()
        QtWidgets.QApplication.processEvents()
        return progress_dialog

    def automatic_refinement(self):
        """
        Refines the current calibration parameters by searching peaks in the approximate positions and subsequent
        refinement. Parameters for this search are set in the GUI.
        """

        # Basic Algorithm:
        # search peaks on first and second ring
        #   calibrate based on those two rings
        #   repeat until ring_ind = max_ind:
        #       search next ring
        #       calibrate based on all previous found points

        num_rings = self.widget.options_num_rings_sb.value()

        progress_dialog = self.create_progress_dialog(
            "Refining Calibration.", "Abort", num_rings
        )
        self.clear_peaks()
        self.load_calibrant(wavelength_from="pyFAI")  # load right calibration file

        # get options
        algorithm = str(self.widget.options_peaksearch_algorithm_cb.currentText())
        delta_tth = float(self.widget.options_delta_tth_txt.text())
        intensity_min_factor = float(
            self.widget.options_intensity_mean_factor_sb.value()
        )
        intensity_max = float(self.widget.options_intensity_limit_txt.text())

        self.model.calibration_model.setup_peak_search_algorithm(algorithm)

        if self.widget.use_mask_cb.isChecked():
            mask = self.model.mask_model.get_img()
        else:
            mask = None

        self.model.calibration_model.search_peaks_on_ring(
            0, delta_tth, intensity_min_factor, intensity_max, mask
        )
        self.widget.peak_num_sb.setValue(2)
        progress_dialog.setValue(1)
        self.model.calibration_model.search_peaks_on_ring(
            1, delta_tth, intensity_min_factor, intensity_max, mask
        )
        self.widget.peak_num_sb.setValue(3)
        if len(self.model.calibration_model.points):
            self.model.calibration_model.refine()
            self.plot_points()
        else:
            print(
                "Did not find any Points with the specified parameters for the first two rings!"
            )

        progress_dialog.setValue(2)

        refinement_canceled = False
        for i in range(num_rings - 2):
            try:
                points = self.model.calibration_model.search_peaks_on_ring(
                    i + 2, delta_tth, intensity_min_factor, intensity_max, mask
                )
            except NotEnoughSpacingsInCalibrant:
                QtWidgets.QMessageBox.critical(
                    self.widget,
                    "Not enough d-spacings!.",
                    "The calibrant file does not contain enough d-spacings.",
                    QtWidgets.QMessageBox.Ok,
                )
                break
            self.widget.peak_num_sb.setValue(i + 4)
            if len(self.model.calibration_model.points):
                self.plot_points(points)
                QtWidgets.QApplication.processEvents()
                QtWidgets.QApplication.processEvents()
                self.model.calibration_model.refine()
            else:
                print("Did not find enough points with the specified parameters!")
            progress_dialog.setLabelText(
                "Refining Calibration. \n" "Finding peaks on Ring {0}.".format(i + 3)
            )
            progress_dialog.setValue(i + 3)
            if progress_dialog.wasCanceled():
                refinement_canceled = True
                break
        progress_dialog.close()
        del progress_dialog

        QtWidgets.QApplication.processEvents()
        if len(self.model.calibration_model.points) == 0:
            QtWidgets.QMessageBox.critical(
                self.widget,
                "No peaks found!",
                "Automatic peak search did not find any peaks.\nThis might be due to the peak search settings or the wrong calibrant being selected.",
                QtWidgets.QMessageBox.Ok,
            )
        if not refinement_canceled and len(self.model.calibration_model.points) > 0:
            self.update_all()

    def load_calibration(self):
        """
        Loads a '*.poni' file and updates the calibration data class
        """
        filename = open_file_dialog(
            self.widget,
            caption="Load calibration...",
            directory=self.model.working_directories["calibration"],
            filter="*.poni",
        )
        if filename != "":
            self.model.working_directories["calibration"] = os.path.dirname(filename)
            self.model.calibration_model.load(filename)
            self.update_all(integrate=self.model.img_model.filename != "")

    def plot_mask(self):
        """
        Plots the mask
        """
        state = self.widget.use_mask_cb.isChecked()
        if state:
            self.widget.img_widget.activate_mask()
            self.widget.img_widget.plot_mask(self.model.mask_model.get_img())
        else:
            self.widget.img_widget.deactivate_mask()

    def mask_transparent_status_changed(self, state):
        """
        :param state: Boolean value whether the mask is being transparent
        :type state: bool
        """
        if state:
            self.widget.img_widget.set_mask_color([255, 0, 0, 100])
        else:
            self.widget.img_widget.set_mask_color([255, 0, 0, 255])

    def update_all(self, integrate=True):
        """
        Performs 1d and 2d integration based on the current calibration parameter set. Updates the GUI interface
        accordingly with the new diffraction pattern and cake image.
        """
        if integrate:
            progress_dialog = self.create_progress_dialog(
                "Integrating to cake.", "", 0, show_cancel_btn=False
            )
            QtWidgets.QApplication.processEvents()
            self.model.current_configuration.integrate_image_2d()
            progress_dialog.setLabelText("Integrating to pattern.")
            QtWidgets.QApplication.processEvents()
            QtWidgets.QApplication.processEvents()
            self.model.current_configuration.integrate_image_1d()
            progress_dialog.close()
        self.widget.cake_widget.plot_image(self.model.cake_data, False)
        self.widget.cake_widget.auto_level()

        self.widget.pattern_widget.plot_data(*self.model.pattern.data)
        self.widget.pattern_widget.plot_vertical_lines(
            self.convert_x_value(
                np.array(self.model.calibration_model.calibrant.get_2th())
                / np.pi
                * 180,
                "2th_deg",
                self.model.current_configuration.integration_unit,
                None,
            )
        )

        if self.model.current_configuration.integration_unit == "2th_deg":
            self.widget.pattern_widget.pattern_plot.setLabel("bottom", "2θ", "°")
        elif self.model.current_configuration.integration_unit == "q_A^-1":
            self.widget.pattern_widget.pattern_plot.setLabel(
                "bottom", "Q", "A<sup>-1</sup>"
            )
        elif self.model.current_configuration.integration_unit == "d_A":
            self.widget.pattern_widget.pattern_plot.setLabel("bottom", "d", "A")

        self.widget.pattern_widget.view_box.autoRange()
        if self.widget.tab_widget.currentIndex() == 0:
            self.widget.tab_widget.setCurrentIndex(1)

        if (
            self.widget.ToolBox.currentIndex() != 2
            or self.widget.ToolBox.currentIndex() != 3
        ):
            self.widget.ToolBox.setCurrentIndex(2)
        self.update_calibration_parameter_in_view()
        self.load_calibrant("pyFAI")

    def update_calibration_parameter_in_view(self):
        """
        Reads the calibration parameter from the calibration_data object and displays them in the GUI.
        :return:
        """
        pyFAI_parameter, fit2d_parameter = (
            self.model.calibration_model.get_calibration_parameter()
        )
        self.widget.set_calibration_parameters(pyFAI_parameter, fit2d_parameter)
        self._update_spline_in_gui()

    def update_detector_parameters_in_view(self):
        detector_mode = self.model.calibration_model.detector_mode

        self.widget.enable_pixel_size_txt(detector_mode == DetectorModes.CUSTOM)
        self.widget.detectors_cb.setVisible(
            detector_mode in (DetectorModes.CUSTOM, DetectorModes.PREDEFINED)
        )
        self.widget.detector_name_lbl.setVisible(detector_mode == DetectorModes.NEXUS)
        self.widget.detector_reset_btn.setEnabled(detector_mode == DetectorModes.NEXUS)

        if detector_mode == DetectorModes.CUSTOM:
            self.widget.detectors_cb.blockSignals(True)
            self.widget.detectors_cb.setCurrentText("Custom")
            self.widget.detectors_cb.blockSignals(False)

        if detector_mode == DetectorModes.PREDEFINED:
            self.widget.detectors_cb.blockSignals(True)
            self.widget.detectors_cb.setCurrentText(
                self.model.calibration_model.detector.name
            )
            self.widget.detectors_cb.blockSignals(False)

        if detector_mode == DetectorModes.NEXUS:
            self.widget.detector_name_lbl.setText(
                os.path.basename(self.model.calibration_model.detector.filename)
            )

        self._update_pixel_size_in_gui()
        self._update_spline_in_gui()

    def show_detector_reset_message_box(self):
        QtWidgets.QMessageBox.critical(
            self.widget,
            "Shape mismatch.",
            "Image and detector definition do not have the same shape!\n"
            + "The Detector has been reset.",
            QtWidgets.QMessageBox.Ok,
        )

    def save_calibration(self):
        """
        Saves the current calibration in a file.
        :return:
        """

        filename = save_file_dialog(
            self.widget,
            "Save calibration...",
            self.model.working_directories["calibration"],
            "*.poni",
        )
        if filename != "":
            self.model.working_directories["calibration"] = os.path.dirname(filename)
            if not filename.rsplit(".", 1)[-1] == "poni":
                filename = filename + ".poni"
            self.model.calibration_model.save(filename)

    def update_img_mouse_position_lbl(self, x, y):
        """
        Displays the values of x, y (usually mouse -position) and their image intensity in the GUI.
        """
        # x, y = pos
        try:
            if x > 0 and y > 0:
                str = "x: %.1f y: %.1f I: %.0f" % (
                    x,
                    y,
                    self.widget.img_widget.img_data.T[
                        int(np.round(x)), int(np.round(y))
                    ],
                )
            else:
                str = "x: %.1f y: %.1f" % (x, y)
        except (IndexError, AttributeError):
            str = "x: %.1f y: %.1f" % (x, y)
        self.widget.pos_lbl.setText(str)

    def update_cake_mouse_position_lbl(self, x, y):
        """
        Displays the values of x, y (usually mouse -position) and their cake intensity in the GUI.
        """
        # x, y = pos
        try:
            if x > 0 and y > 0:
                str = "x: %.1f y: %.1f I: %.0f" % (
                    x,
                    y,
                    self.widget.cake_widget.img_data.T[
                        int(np.round(x)), int(np.round(y))
                    ],
                )
            else:
                str = "x: %.1f y: %.1f" % (x, y)
        except (IndexError, AttributeError):
            str = "x: %.1f y: %.1f" % (x, y)
        self.widget.pos_lbl.setText(str)

    def update_pattern_mouse_position_lbl(self, x, y):
        """
        Displays the values of x, y (pattern mouse-position) in the GUI.
        """
        # x, y = pos
        str = "x: %.1f y: %.1f" % (x, y)
        self.widget.pos_lbl.setText(str)

    def convert_x_value(self, value, previous_unit, new_unit, wavelength):
        if wavelength is None:
            wavelength = self.model.calibration_model.wavelength
        if previous_unit == "2th_deg":
            tth = value
        elif previous_unit == "q_A^-1":
            tth = np.arcsin(value * 1e10 * wavelength / (4 * np.pi)) * 360 / np.pi
        elif previous_unit == "d_A":
            tth = 2 * np.arcsin(wavelength / (2 * value * 1e-10)) * 180 / np.pi
        else:
            tth = 0

        if new_unit == "2th_deg":
            res = tth
        elif new_unit == "q_A^-1":
            res = 4 * np.pi * np.sin(tth / 360 * np.pi) / wavelength / 1e10
        elif new_unit == "d_A":
            res = wavelength / (2 * np.sin(tth / 360 * np.pi)) * 1e10
        else:
            res = 0
        return res



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from qtpy import QtWidgets, QtCore

from ...CustomWidgets import (
    HorizontalSpacerItem,
    MenuTabWidget,
    NumberTextField,
    ListTableWidget,
    CheckableButton,
    CheckableFlatButton,
    VerticalSpacerItem,
)


class CorrectionsWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(CorrectionsWidget, self).__init__(*args, **kwargs)

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 5, 0, 0)

        self.create_cbn_correction_widgets()
        self.create_cbn_correction_layout()

        self.create_oiadac_widgets()
        self.create_oiadac_layout()

        self.create_transfer_widgets()
        self.create_transfer_layout()

        self.setLayout(self._layout)

        self.menu_tab_widget = MenuTabWidget()
        self.menu_tab_widget.add_tab("cBN Seat", self.cbn_seat_gb)
        self.menu_tab_widget.add_tab("Inc. Abs.", self.oiadac_gb)
        self.menu_tab_widget.add_tab("Transfer", self.transfer_gb)
        self.menu_tab_widget.select_tab(0)

        self._layout.addWidget(self.menu_tab_widget)

        self.style_widgets()

    def create_cbn_correction_widgets(self):
        self.cbn_seat_gb = QtWidgets.QGroupBox("cBN Seat Correction")
        self.cbn_seat_plot_btn = CheckableButton("Plot")

        self.cbn_param_tw = ListTableWidget()
        self.cbn_param_tw.setColumnCount(3)

        self.cbn_param_tw.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.Stretch
        )
        self.cbn_param_tw.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)

        cbn_parameters = [
            ["Anvil thickness", 2.3, "mm"],
            ["Seat thickness", 5.3, "mm"],
            ["Inner seat radius", 0.4, "mm"],
            ["Outer seat radius", 1.95, "mm"],
            ["Cell tilt", 0.0, "°"],
            ["Cell tilt rotation", 0, "°"],
            ["Center offset", 0, "mm"],
            ["Center offset rotation", 0, "°"],
            ["Anvil absorption length", 13.7, "mm"],
            ["Seat absorption length", 12, "mm"],
        ]

        for cbn_parameter in cbn_parameters:
            self.add_param_to_tw(self.cbn_param_tw, *cbn_parameter)

    @staticmethod
    def add_param_to_tw(tw, name, value, unit):
        tw.blockSignals(True)
        new_row_ind = int(tw.rowCount())
        tw.setRowCount(new_row_ind + 1)

        name_item = QtWidgets.QTableWidgetItem(name + ":")
        name_item.setFlags(name_item.flags() & ~QtCore.Qt.ItemIsEditable)
        name_item.setTextAlignment(int(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter))
        tw.setItem(new_row_ind, 0, name_item)

        value_item = NumberTextField("{:g}".format(value))
        tw.setCellWidget(new_row_ind, 1, value_item)

        unit_item = QtWidgets.QTableWidgetItem(unit)
        unit_item.setFlags(name_item.flags() & ~QtCore.Qt.ItemIsEditable)
        unit_item.setTextAlignment(int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter))
        tw.setItem(new_row_ind, 2, unit_item)

        tw.resizeColumnToContents(0)
        tw.resizeColumnToContents(2)

        tw.blockSignals(False)

    def create_cbn_correction_layout(self):
        self._cbn_seat_layout = QtWidgets.QHBoxLayout()
        self._cbn_seat_layout.setSpacing(5)

        self._cbn_seat_layout.addWidget(self.cbn_param_tw)

        self._cbn_seat_right_layout = QtWidgets.QVBoxLayout()
        self._cbn_seat_right_layout.addWidget(self.cbn_seat_plot_btn)
        self._cbn_seat_right_layout.addStretch()
        self._cbn_seat_layout.addLayout(self._cbn_seat_right_layout)

        self.cbn_seat_gb.setLayout(self._cbn_seat_layout)

    def create_oiadac_widgets(self):
        self.oiadac_gb = QtWidgets.QGroupBox("Detector Incidence Absorption Correction")

        self.oiadac_param_tw = ListTableWidget()
        self.oiadac_param_tw.setColumnCount(3)

        self.oiadac_param_tw.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.Stretch
        )
        self.oiadac_param_tw.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)

        self.detector_thickness_txt = NumberTextField("40")
        self.detector_absorption_length_txt = NumberTextField("465.5")

        oiadac_parameters = [
            ["Detector thickness", 40, "mm"],
            ["Detector absorption length", 465.5, "um"],
        ]

        for param in oiadac_parameters:
            self.add_param_to_tw(self.oiadac_param_tw, *param)

        self.oiadac_plot_btn = CheckableButton("Plot")

    def create_oiadac_layout(self):
        self._oiadac_layout = QtWidgets.QHBoxLayout()
        self._oiadac_layout.setSpacing(5)

        self._oiadac_layout.addWidget(self.oiadac_param_tw)

        self._oiadac_right_layout = QtWidgets.QVBoxLayout()
        self._oiadac_right_layout.addWidget(self.oiadac_plot_btn)
        self._oiadac_right_layout.addStretch()
        self._oiadac_layout.addLayout(self._oiadac_right_layout)

        self.oiadac_gb.setLayout(self._oiadac_layout)

    def create_transfer_widgets(self):
        self.transfer_gb = QtWidgets.QGroupBox("Transfer Correction")
        self.transfer_load_original_btn = QtWidgets.QPushButton("Load Original")
        self.transfer_load_response_btn = QtWidgets.QPushButton("Load Response")
        self.transfer_original_filename_lbl = QtWidgets.QLabel("None")
        self.transfer_response_filename_lbl = QtWidgets.QLabel("None")
        self.transfer_plot_btn = CheckableButton("Plot")

    def create_transfer_layout(self):
        self._transfer_layout = QtWidgets.QGridLayout()
        self._transfer_layout.setSpacing(5)
        self._transfer_layout.addWidget(self.transfer_load_original_btn, 0, 0)
        self._transfer_layout.addWidget(self.transfer_load_response_btn, 1, 0)
        self._transfer_layout.addWidget(self.transfer_original_filename_lbl, 0, 1)
        self._transfer_layout.addWidget(self.transfer_response_filename_lbl, 1, 1)
        self._transfer_layout.addWidget(self.transfer_plot_btn, 0, 2)
        self._transfer_layout.setColumnStretch(0, 0)
        self._transfer_layout.setColumnStretch(1, 1)
        self._transfer_layout.setColumnStretch(2, 0)
        self._transfer_layout.setRowStretch(0, 0)
        self._transfer_layout.setRowStretch(1, 0)
        self._transfer_layout.setRowStretch(2, 1)
        self.transfer_gb.setLayout(self._transfer_layout)

    def style_widgets(self):
        self.cbn_seat_gb.setCheckable(True)
        self.cbn_seat_gb.setChecked(False)
        self.oiadac_gb.setCheckable(True)
        self.oiadac_gb.setChecked(False)
        self.transfer_gb.setCheckable(True)
        self.transfer_gb.setChecked(False)

        self.setStyleSheet(
            """
                    QLineEdit {
                        min-width: 50 px;
                        min-height: 26 px;
                        max-height: 26 px;
                    }
                    """
        )



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pyqtgraph as pg
import numpy as np
from qtpy import QtCore
from pyqtgraph.exporters.ImageExporter import ImageExporter
from pyqtgraph.exporters.SVGExporter import SVGExporter

from xypattern import Pattern

from dioptas.model.OverlayModel import Overlay

from .ExLegendItem import LegendItem


class PatternWidget(QtCore.QObject):
    mouse_moved = QtCore.Signal(float, float)
    mouse_left_clicked = QtCore.Signal(float, float)
    range_changed = QtCore.Signal(list)
    auto_range_status_changed = QtCore.Signal(bool)

    def __init__(self, pg_layout):
        super(PatternWidget, self).__init__()
        self.pg_layout = pg_layout
        self.create_graphics()
        self.create_main_plot()
        self.create_pos_line()
        self.modify_mouse_behavior()

        self._auto_range = True

        self.phases = []  # type: list[PhasePlot]
        self.phases_vlines = []

        self.overlays = []

        self.plot_name = ""

    def create_graphics(self):
        self.pattern_plot = self.pg_layout.addPlot(
            labels={"left": "Intensity", "bottom": "2 Theta"}
        )
        self.pattern_plot.setLabel("bottom", "2θ", "°")
        self.pattern_plot.enableAutoRange(False)
        self.pattern_plot.buttonsHidden = True
        self.view_box = self.pattern_plot.vb
        self.legend = LegendItem(
            horSpacing=20,
            box=False,
            verSpacing=-3,
            labelAlignment="right",
            showLines=False,
        )
        self.phases_legend = LegendItem(
            horSpacing=20,
            box=False,
            verSpacing=-3,
            labelAlignment="left",
            showLines=False,
        )

    def create_main_plot(self):
        self.plot_item = pg.PlotDataItem(
            np.linspace(0, 10),
            np.sin(np.linspace(10, 3)),
            pen=pg.mkPen(color=(255, 255, 255), width=2),
        )
        self.pattern_plot.addItem(self.plot_item)
        self.bkg_item = pg.PlotDataItem(
            [], [], pen=pg.mkPen(color=(255, 0, 0), width=2, style=QtCore.Qt.DashLine)
        )
        self.pattern_plot.addItem(self.bkg_item)
        self.legend.addItem(self.plot_item, "")
        self.legend.setParentItem(self.pattern_plot.vb)
        self.legend.anchor(itemPos=(1, 0), parentPos=(1, 0), offset=(-10, -10))
        self.phases_legend.setParentItem(self.pattern_plot.vb)
        self.phases_legend.anchor(itemPos=(0, 0), parentPos=(0, 0), offset=(0, -10))

        self.bkg_roi = ModifiedLinearRegionItem(
            [5, 20],
            pg.LinearRegionItem.Vertical,
        )
        self.map_interactive_roi = SymmetricModifiedLinearRegionItem(
            [5, 7],
            pg.LinearRegionItem.Vertical,
        )

    @property
    def auto_range(self):
        return self._auto_range

    @auto_range.setter
    def auto_range(self, value):
        if self._auto_range is not value:
            self._auto_range = value
            self.auto_range_status_changed.emit(value)
        if self._auto_range is True:
            self.update_graph_range()

    def create_pos_line(self):
        self.pos_line = pg.InfiniteLine(
            pen=pg.mkPen(color=(0, 255, 0), width=1.5, style=QtCore.Qt.DashLine)
        )
        self.pattern_plot.addItem(self.pos_line)

    def deactivate_pos_line(self):
        self.pattern_plot.removeItem(self.pos_line)

    def set_pos_line(self, x):
        self.pos_line.setPos(x)

    def get_pos_line(self):
        return self.pos_line.value()

    def plot_data(self, x, y, name=None):
        self.plot_item.setData(x, y)
        if name is not None:
            self.legend.legendItems[0][1].setText(name)
            self.plot_name = name
        self.legend.updateSize()
        self.update_graph_range()

    def plot_bkg(self, x, y):
        self.bkg_item.setData(x, y)

    def update_graph_range(self):
        x_range = list(self.plot_item.dataBounds(0))
        y_range = list(self.plot_item.dataBounds(1))

        for overlay in self.overlays:
            if overlay in self.pattern_plot.items:
                x_range_overlay = overlay.dataBounds(0)
                y_range_overlay = overlay.dataBounds(1)
                if x_range_overlay[0] < x_range[0]:
                    x_range[0] = x_range_overlay[0]
                if x_range_overlay[1] > x_range[1]:
                    x_range[1] = x_range_overlay[1]
                if y_range_overlay[0] < y_range[0]:
                    y_range[0] = y_range_overlay[0]
                if y_range_overlay[1] > y_range[1]:
                    y_range[1] = y_range_overlay[1]

        if x_range[1] is not None and x_range[0] is not None:
            padding = self.view_box.suggestPadding(0)
            diff = x_range[1] - x_range[0]
            x_range = [x_range[0] - padding * diff, x_range[1] + padding * diff]

            self.view_box.setLimits(xMin=x_range[0], xMax=x_range[1])

            if self.auto_range:
                self.view_box.setRange(xRange=x_range, padding=0)

        if y_range[1] is not None and y_range[0] is not None:
            padding = self.view_box.suggestPadding(1)
            diff = y_range[1] - y_range[0]
            y_range = [y_range[0] - padding * diff, y_range[1] + padding * diff]

            self.view_box.setLimits(yMin=y_range[0], yMax=y_range[1])

            if self.auto_range:
                self.view_box.setRange(yRange=y_range, padding=0)
        self.emit_sig_range_changed()

    def add_overlay(
        self, pattern: Pattern, color: tuple[int, int, int], show: bool = True
    ):
        x, y = pattern.data
        self.overlays.append(
            pg.PlotDataItem(x, y, pen=pg.mkPen(color=color, width=1.5))
        )
        if show:
            self.pattern_plot.addItem(self.overlays[-1])
            self.legend.addItem(self.overlays[-1], pattern.name)
            self.update_graph_range()
        return color

    def remove_overlay(self, ind):
        self.pattern_plot.removeItem(self.overlays[ind])
        self.legend.removeItem(self.overlays[ind])
        self.overlays.remove(self.overlays[ind])

    def hide_overlay(self, ind):
        if not self.overlays[ind] in self.pattern_plot.items:
            return
        self.pattern_plot.removeItem(self.overlays[ind])
        self.legend.hideItem(ind + 1)

    def show_overlay(self, ind):
        if self.overlays[ind] in self.pattern_plot.items:
            return
        self.pattern_plot.addItem(self.overlays[ind])
        self.legend.showItem(ind + 1)

    def update_overlay(self, ind: int, overlay: Overlay):
        self.set_overlay_data(ind, overlay.x, overlay.y)
        self.set_overlay_color(ind, overlay.color)
        self.rename_overlay(ind, overlay.name)
        if overlay.visible:
            self.show_overlay(ind)
        else:
            self.hide_overlay(ind)
        self.update_graph_range()

    def set_overlay_data(self, ind, x, y):
        self.overlays[ind].setData(x, y)

    def set_overlay_color(self, ind: int, color):
        self.overlays[ind].setPen(pg.mkPen(color=color, width=1.5))
        self.legend.setItemColor(ind + 1, color)

    def rename_overlay(self, ind: int, name: str):
        self.legend.renameItem(ind + 1, name)

    def set_antialias(self, value):
        for overlay in self.overlays:
            overlay.opts["antialias"] = value
            overlay.updateItems()

    def add_phase(self, name, positions, intensities, baseline, color):
        self.phases.append(
            PhasePlot(
                self.pattern_plot,
                self.phases_legend,
                positions,
                intensities,
                color,
                name,
                baseline,
            )
        )

    def set_phase_color(self, ind, color):
        self.phases[ind].set_color(color)
        self.phases_legend.setItemColor(ind, color)

    def hide_phase(self, ind):
        self.phases[ind].hide()
        self.phases_legend.hideItem(ind)

    def show_phase(self, ind):
        self.phases[ind].show()
        self.phases_legend.showItem(ind)

    def rename_phase(self, ind, name):
        self.phases_legend.renameItem(ind, name)

    def update_phase_intensities(self, ind, positions, intensities, baseline=0):
        if len(self.phases):
            self.phases[ind].update_intensities(positions, intensities, baseline)

    def update_phase_line_visibility(self, ind):
        x_range = self.plot_item.dataBounds(0)
        self.phases[ind].update_visibilities(x_range)

    def update_phase_line_visibilities(self):
        x_range = self.plot_item.dataBounds(0)
        for phase in self.phases:
            phase.update_visibilities(x_range)

    def del_phase(self, ind):
        self.phases[ind].remove()
        del self.phases[ind]

    def plot_vertical_lines(self, positions, name=None):
        if len(self.phases_vlines) > 0:
            self.phases_vlines[0].set_data(positions, name)
        else:
            self.phases_vlines.append(
                PhaseLinesPlot(
                    self.pattern_plot,
                    positions,
                    pen=pg.mkPen(color=(200, 50, 50), style=QtCore.Qt.SolidLine),
                )
            )

    # bkg_linear region functions
    def show_bkg_roi(self):
        self.pattern_plot.addItem(self.bkg_roi)

    def set_bkg_roi(self, x_min, x_max):
        self.bkg_roi.setRegion((x_min, x_max))

    def get_bkg_roi(self):
        return self.bkg_roi.getRegion()

    def hide_bkg_roi(self):
        self.pattern_plot.removeItem(self.bkg_roi)

    # map interactive linear region functions
    def show_map_interactive_roi(self):
        self.pattern_plot.addItem(self.map_interactive_roi)

    def set_map_interactive_roi(self, x_min, x_max):
        self.map_interactive_roi.setRegion((x_min, x_max))

    def hide_map_interactive_roi(self):
        self.pattern_plot.removeItem(self.map_interactive_roi)

    # saving the plot
    def save_png(self, filename):
        exporter = ImageExporter(self.pattern_plot)
        exporter.export(filename)

    def save_svg(self, filename):
        self._invert_color()
        previous_label = None
        if self.pattern_plot.getAxis("bottom").labelText == "2θ":
            previous_label = ("2θ", "°")
            self.pattern_plot.setLabel("bottom", "2th_deg", "")
        exporter = SVGExporter(self.pattern_plot)
        exporter.export(filename)
        self._norm_color()
        if previous_label is not None:
            self.pattern_plot.setLabel("bottom", previous_label[0], previous_label[1])

    def _invert_color(self):
        self.pattern_plot.getAxis("bottom").setPen("k")
        self.pattern_plot.getAxis("left").setPen("k")
        self.plot_item.setPen("k")
        self.legend.legendItems[0][1].setText(
            self.legend.legendItems[0][1].text, color="#000"
        )

    def _norm_color(self):
        self.pattern_plot.getAxis("bottom").setPen("w")
        self.pattern_plot.getAxis("left").setPen("w")
        self.plot_item.setPen("w")
        self.legend.legendItems[0][1].setText(
            self.legend.legendItems[0][1].text, color="#FFF"
        )

    def mouseMoved(self, pos):
        pos = self.plot_item.mapFromScene(pos)
        self.mouse_moved.emit(pos.x(), pos.y())

    def modify_mouse_behavior(self):
        # different mouse handlers
        self.view_box.setMouseMode(self.view_box.RectMode)

        self.pg_layout.scene().sigMouseMoved.connect(self.mouseMoved)
        self.view_box.mouseClickEvent = self.myMouseClickEvent
        self.view_box.mouseDragEvent = self.myMouseDragEvent
        self.view_box.mouseDoubleClickEvent = self.myMouseDoubleClickEvent
        self.view_box.wheelEvent = self.myWheelEvent

        # create sigranged changed timer for right click drag
        # if not using the timer the signals are fired too often and
        # the computer becomes slow...
        self.range_changed_timer = QtCore.QTimer()
        self.range_changed_timer.timeout.connect(self.emit_sig_range_changed)
        self.range_changed_timer.setInterval(30)
        self.last_view_range = np.array(self.view_box.viewRange())

    def myMouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.RightButton or (
            ev.button() == QtCore.Qt.LeftButton
            and ev.modifiers() & QtCore.Qt.ControlModifier
        ):
            view_range = np.array(self.view_box.viewRange()) * 2
            curve_data = self.plot_item.getData()
            x_range = np.max(curve_data[0]) - np.min(curve_data[0])
            if (view_range[0][1] - view_range[0][0]) > x_range:
                self.auto_range = True
            else:
                self.auto_range = False
                self.view_box.scaleBy((2, 2))
            self.emit_sig_range_changed()
        elif ev.button() == QtCore.Qt.LeftButton:
            pos = self.view_box.mapFromScene(ev.pos())
            pos = self.plot_item.mapFromScene(2 * ev.pos() - pos)
            x = pos.x()
            y = pos.y()
            self.mouse_left_clicked.emit(x, y)

    def myMouseDoubleClickEvent(self, ev):
        if (ev.button() == QtCore.Qt.RightButton) or (
            ev.button() == QtCore.Qt.LeftButton
            and ev.modifiers() & QtCore.Qt.ControlModifier
        ):
            self.auto_range = True
            self.emit_sig_range_changed()

    def myMouseDragEvent(self, ev, axis=None):
        # most of this code is copied behavior mouse drag from the original code
        ev.accept()
        pos = ev.pos()
        lastPos = ev.lastPos()
        dif = pos - lastPos
        dif *= -1

        if ev.button() == QtCore.Qt.RightButton or (
            ev.button() == QtCore.Qt.LeftButton
            and ev.modifiers() & QtCore.Qt.ControlModifier
        ):
            # determine the amount of translation
            tr = dif
            tr = self.view_box.mapToView(tr) - self.view_box.mapToView(pg.Point(0, 0))
            x = tr.x()
            y = tr.y()
            self.view_box.translateBy(x=x, y=y)
            if ev.start:
                self.range_changed_timer.start()
            if ev.isFinish():
                self.range_changed_timer.stop()
                self.emit_sig_range_changed()
        else:
            if (
                ev.isFinish()
            ):  # This is the final move in the drag; change the view scale now
                self.auto_range = False
                self.view_box.rbScaleBox.hide()
                ax = QtCore.QRectF(
                    pg.Point(ev.buttonDownPos(ev.button())), pg.Point(pos)
                )
                ax = self.view_box.childGroup.mapRectFromParent(ax)
                self.view_box.showAxRect(ax)
                self.view_box.axHistoryPointer += 1
                self.view_box.axHistory = self.view_box.axHistory[
                    : self.view_box.axHistoryPointer
                ] + [ax]
                self.emit_sig_range_changed()
            else:
                # update shape of scale box
                self.view_box.updateScaleBox(ev.buttonDownPos(), ev.pos())

    def emit_sig_range_changed(self):
        new_view_range = np.array(self.view_box.viewRange())
        if not np.array_equal(self.last_view_range, new_view_range):
            self.view_box.sigRangeChangedManually.emit(
                self.view_box.state["mouseEnabled"]
            )
            self.last_view_range = new_view_range

    def myWheelEvent(self, ev, axis=None, *args):
        if ev.delta() > 0:
            pg.ViewBox.wheelEvent(self.view_box, ev, axis)

            self.auto_range = False
            # axis_range = self.pattern_plot.viewRange()
            # self.range_changed.emit(axis_range)
            self.emit_sig_range_changed()
        else:
            if self.auto_range is not True:
                view_range = np.array(self.view_box.viewRange())
                curve_data = self.plot_item.getData()
                self.auto_range = False
                if len(curve_data[0]) > 2:
                    x_range = np.max(curve_data[0]) - np.min(curve_data[0])
                    y_range = np.max(curve_data[1]) - np.min(curve_data[1])
                    if (view_range[0][1] - view_range[0][0]) >= x_range and (
                        view_range[1][1] - view_range[1][0]
                    ) >= y_range:
                        self.auto_range = True
                if not self.auto_range:
                    pg.ViewBox.wheelEvent(self.view_box, ev)
            self.emit_sig_range_changed()


class PhaseLinesPlot(object):
    def __init__(
        self,
        plot_item,
        positions=None,
        name="Dummy",
        pen=pg.mkPen(color=(120, 120, 120), style=QtCore.Qt.DashLine),
    ):
        self.plot_item = plot_item
        self.peak_positions = []
        self.line_items = []
        self.pen = pen
        self.name = name

        if positions is not None:
            self.set_data(positions, name)

    def set_data(self, positions, name):
        # remove all old lines
        for item in self.line_items:
            self.plot_item.removeItem(item)

        # create new ones on each Position:
        self.line_items = []
        self.peak_positions = positions
        for ind, position in enumerate(positions):
            self.line_items.append(pg.InfiniteLine(pen=self.pen))
            self.line_items[ind].setValue(position)
            self.plot_item.addItem(self.line_items[ind])


class PhasePlot(object):
    num_phases = 0

    def __init__(
        self,
        plot_item,
        legend_item,
        positions,
        intensities,
        color,
        name=None,
        baseline=0,
    ):
        self.plot_item = plot_item
        self.legend_item = legend_item
        self.visible = True
        self.line_items = []  # type: list[pg.PlotDataItem]
        self.line_visible = []
        self.pattern_x_range = []
        self.index = PhasePlot.num_phases
        self.color = color
        self.pen = pg.mkPen(color=self.color, width=0.9, style=QtCore.Qt.SolidLine)
        self.ref_legend_line = pg.PlotDataItem(pen=self.pen)
        self.name = ""
        PhasePlot.num_phases += 1
        self.create_items(positions, intensities, name, baseline)

    def create_items(self, positions, intensities, name=None, baseline=0):
        # create new ones on each Position:
        self.line_items = []

        for ind, position in enumerate(positions):
            self.line_items.append(
                pg.PlotDataItem(
                    x=[position, position],
                    y=[baseline, intensities[ind]],
                    pen=self.pen,
                    antialias=False,
                )
            )
            self.line_visible.append(True)
            self.plot_item.addItem(self.line_items[ind])

        if name is not None:
            try:
                self.legend_item.addItem(self.ref_legend_line, name)
                self.name = name
            except IndexError:
                pass

    def add_line(self):
        self.line_items.append(
            pg.PlotDataItem(x=[0, 0], y=[0, 0], pen=self.pen, antialias=False)
        )
        self.line_visible.append(True)
        self.plot_item.blockSignals(True)
        self.plot_item.addItem(self.line_items[-1])
        self.plot_item.blockSignals(False)

    def delete_line(self, ind=-1):
        self.plot_item.removeItem(self.line_items[ind])
        del self.line_items[ind]
        del self.line_visible[ind]

    def clear_lines(self):
        for dummy_ind in range(len(self.line_items)):
            self.delete_line()

    def update_intensities(self, positions, intensities, baseline=0):
        if self.visible:
            for ind, intensity in enumerate(intensities):
                self.line_items[ind].setData(
                    y=[baseline, intensity], x=[positions[ind], positions[ind]]
                )

    def update_visibilities(self, pattern_range):
        if self.visible and pattern_range[0] is not None:
            for ind, line_item in enumerate(self.line_items):
                data = line_item.getData()
                position = data[0][0]
                if position >= pattern_range[0] and position <= pattern_range[1]:
                    if not self.line_visible[ind]:
                        self.plot_item.addItem(line_item)
                        self.line_visible[ind] = True
                else:
                    if self.line_visible[ind]:
                        self.plot_item.removeItem(line_item)
                        self.line_visible[ind] = False

    def set_color(self, color):
        self.pen = pg.mkPen(color=color, width=1.3, style=QtCore.Qt.SolidLine)
        for line_item in self.line_items:
            line_item.setPen(self.pen)
        self.ref_legend_line.setPen(self.pen)

    def hide(self):
        if self.visible:
            self.visible = False
            for ind, line_item in enumerate(self.line_items):
                if self.line_visible[ind]:
                    self.plot_item.removeItem(line_item)

    def show(self):
        if not self.visible:
            self.visible = True
            for ind, line_item in enumerate(self.line_items):
                if self.line_visible[ind]:
                    self.plot_item.addItem(line_item)

    def remove(self):
        try:
            self.legend_item.removeItem(self.ref_legend_line)
        except IndexError:
            print("this phase had now lines in the appropriate region")
        for ind, item in enumerate(self.line_items):
            if self.line_visible[ind]:
                self.plot_item.removeItem(item)


class ModifiedLinearRegionItem(pg.LinearRegionItem):
    def __init__(self, *args, **kwargs):
        super(ModifiedLinearRegionItem, self).__init__(*args, **kwargs)

    def mouseDragEvent(self, ev):
        return

    def hoverEvent(self, ev):
        return


class SymmetricModifiedLinearRegionItem(pg.LinearRegionItem):
    def __init__(self, *args, **kwargs):
        super(SymmetricModifiedLinearRegionItem, self).__init__(*args, **kwargs)
        self.center = (
            self.lines[1].getXPos() - self.lines[0].getXPos()
        ) / 2 + self.lines[0].getXPos()

    def setCenter(self, center):
        previous_width = self.lines[1].value() - self.lines[0].value()
        new_left = center - previous_width / 2
        new_right = center + previous_width / 2
        self.center = center
        self.setRegion([new_left, new_right])



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2013-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import time

import os
from qtpy import QtCore, QtWidgets, QtGui
from math import floor, log10

from .. import icons_path


class NumberTextField(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super(NumberTextField, self).__init__(*args, **kwargs)
        self.setValidator(QtGui.QDoubleValidator())
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

    def text(self):
        return super(NumberTextField, self).text().replace(",", ".")

    def value(self):
        return float(self.text())


class IntegerTextField(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super(IntegerTextField, self).__init__(*args, **kwargs)
        self.setValidator(QtGui.QIntValidator())
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)


class LabelAlignRight(QtWidgets.QLabel):
    def __init__(self, *args, **kwargs):
        super(LabelAlignRight, self).__init__(*args, **kwargs)
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)


class LabelExpandable(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super(LabelExpandable, self).__init__(*args, **kwargs)
        self.setStyleSheet(
            """
            color: #F1F1F1;
            background: #3C3C3C;
        """
        )
        self.setReadOnly(True)


class CleanLooksComboBox(QtWidgets.QComboBox):
    cleanlooks = QtWidgets.QStyleFactory.create("motif")

    def __init__(self, *args, **kwargs):
        super(CleanLooksComboBox, self).__init__(*args, **kwargs)
        self.setStyle(CleanLooksComboBox.cleanlooks)
        self.setLineEdit(CleanLooksLineEdit())
        self.lineEdit().clicked.connect(self.showPopup)
        self.popup_closed_time = time.time()

    def showPopup(self):
        if time.time() - self.popup_closed_time > 0.01:
            # prevents showing popup immediately after closing by clicking onto lineEdit.
            super(CleanLooksComboBox, self).showPopup()

    def hidePopup(self):
        super(CleanLooksComboBox, self).hidePopup()
        self.popup_closed_time = time.time()


class CleanLooksLineEdit(QtWidgets.QLineEdit):
    clicked = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super(CleanLooksLineEdit, self).__init__(*args, **kwargs)
        self.installEventFilter(self)
        self.setReadOnly(True)
        self.setStyleSheet(
            """
                margin: 2px; 
                background: #3C3C3C;
            """
        )

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.MouseButtonPress:
            self.clicked.emit()
            return True
        if event.type() == QtCore.QEvent.MouseMove:
            return True
        return super(CleanLooksLineEdit, self).eventFilter(obj, event)


class SpinBoxAlignRight(QtWidgets.QSpinBox):
    def __init__(self, *args, **kwargs):
        super(SpinBoxAlignRight, self).__init__(*args, **kwargs)
        self.setAlignment(QtCore.Qt.AlignRight)


class DoubleSpinBoxAlignRight(QtWidgets.QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        super(DoubleSpinBoxAlignRight, self).__init__(*args, **kwargs)
        self.setAlignment(QtCore.Qt.AlignRight)


class DoubleMultiplySpinBoxAlignRight(QtWidgets.QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        super(DoubleMultiplySpinBoxAlignRight, self).__init__(*args, **kwargs)
        self.setAlignment(QtCore.Qt.AlignRight)

    def stepBy(self, p_int):
        self.setValue(self.calc_new_step(self.value(), p_int))

    def calc_new_step(self, value, p_int):
        pow10floor = 10 ** floor(log10(value))
        if p_int > 0:
            if value / pow10floor < 1.9:
                return pow10floor * 2.0
            elif value / pow10floor < 4.9:
                return pow10floor * 5.0
            else:
                return pow10floor * 10.0
        else:
            if value / pow10floor < 1.1:
                return pow10floor / 2.0
            elif value / pow10floor < 2.1:
                return pow10floor
            elif value / pow10floor < 5.1:
                return pow10floor * 2.0
            else:
                return pow10floor * 5.0


class ConservativeSpinBox(QtWidgets.QSpinBox):
    """
    This is a modification of the QSpinBox class. The ConservativeSpinbox does not emit the valueChanged signal for
    every keypress in the lineedit. The signal is only emitted for the following occasions:
      - pressing enter
      - the spinbox loses focus
      - pressing the up or down arrows
    Also the wheel events are disabled.

    This Spinbox is intended for usage with applications were the change in the spinbox value causes long calculations
    and does a valueChanged signal on every keypress results in a strange behavior.
    """

    valueChanged = QtCore.Signal()

    def __init__(self):
        super(QtWidgets.QSpinBox, self).__init__()

        self.lineEdit().editingFinished.connect(self.valueChanged)
        self.lineEdit().setAlignment(QtCore.Qt.AlignRight)

    def mousePressEvent(self, e: QtGui.QMouseEvent):
        opt = QtWidgets.QStyleOptionSpinBox()
        self.initStyleOption(opt)

        if (
            self.style()
            .subControlRect(
                QtWidgets.QStyle.CC_SpinBox, opt, QtWidgets.QStyle.SC_SpinBoxUp
            )
            .contains(e.pos())
        ):
            self.setValue(self.value() + 1)
            self.valueChanged.emit()
        elif (
            self.style()
            .subControlRect(
                QtWidgets.QStyle.CC_SpinBox, opt, QtWidgets.QStyle.SC_SpinBoxDown
            )
            .contains(e.pos())
        ):
            self.setValue(self.value() - 1)
            self.valueChanged.emit()

    def wheelEvent(self, e: QtGui.QWheelEvent):
        pass


class FlatButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super(FlatButton, self).__init__(*args)
        self.setFlat(True)

    def setHeight(self, height):
        self.setMinimumHeight(height)
        self.setMaximumHeight(height)

    def setWidth(self, width):
        self.setMinimumWidth(width)
        self.setMaximumWidth(width)


class CheckableButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super(CheckableButton, self).__init__(*args)
        self.setCheckable(True)


class CheckableFlatButton(FlatButton):
    def __init__(self, *args):
        super(CheckableFlatButton, self).__init__(*args)
        self.setCheckable(True)


class DarkCheckableFlatButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super(DarkCheckableFlatButton, self).__init__(*args)
        self.setObjectName("dark_checkable_flat_btn")
        self.setCheckable(True)
        self.setFlat(True)

    def setHeight(self, height):
        self.setMinimumHeight(height)
        self.setMaximumHeight(height)

    def setWidth(self, width):
        self.setMinimumWidth(width)
        self.setMaximumWidth(width)


class RotatedCheckableFlatButton(CheckableFlatButton):
    def __init__(self, *args):
        super(RotatedCheckableFlatButton, self).__init__(*args)

    def paintEvent(self, event):
        painter = QtWidgets.QStylePainter(self)
        painter.rotate(270)
        painter.translate(-1 * self.height(), 0)
        painter.drawControl(QtWidgets.QStyle.CE_PushButton, self.getSyleOptions())

    def minimumSizeHint(self):
        size = super(RotatedCheckableFlatButton, self).minimumSizeHint()
        size.transpose()
        return size

    def sizeHint(self):
        size = super(RotatedCheckableFlatButton, self).sizeHint()
        size.transpose()
        return size

    def getSyleOptions(self):
        options = QtWidgets.QStyleOptionButton()
        options.initFrom(self)
        size = options.rect.size()
        size.transpose()
        options.rect.setSize(size)
        if self.isFlat():
            options.features |= QtWidgets.QStyleOptionButton.Flat
        if self.menu():
            options.features |= QtWidgets.QStyleOptionButton.HasMenu
        if self.autoDefault() or self.isDefault():
            options.features |= QtWidgets.QStyleOptionButton.AutoDefaultButton
        if self.isDefault():
            options.features |= QtWidgets.QStyleOptionButton.DefaultButton
        if self.isDown() or (self.menu() and self.menu().isVisible()):
            options.state |= QtWidgets.QStyle.State_Sunken
        if self.isChecked():
            options.state |= QtWidgets.QStyle.State_On
        if not self.isFlat() and not self.isDown():
            options.state |= QtWidgets.QStyle.State_Raised

        options.text = self.text()
        options.icon = self.icon()
        options.iconSize = self.iconSize()
        return options


class SaveIconButton(FlatButton):
    def __init__(self):
        super(SaveIconButton, self).__init__()
        self.setIcon(QtGui.QIcon(os.path.join(icons_path, "save.ico")))


class OpenIconButton(FlatButton):
    def __init__(self):
        super(OpenIconButton, self).__init__()
        self.setIcon(QtGui.QIcon(os.path.join(icons_path, "open.ico")))


class ResetIconButton(FlatButton):
    def __init__(self):
        super(ResetIconButton, self).__init__()
        self.setIcon(QtGui.QIcon(os.path.join(icons_path, "reset.ico")))


class HorizontalLine(QtWidgets.QFrame):
    def __init__(self):
        super(HorizontalLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setFixedHeight(1)


class VerticalLine(QtWidgets.QFrame):
    def __init__(self):
        super(VerticalLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setFixedWidth(1)


class ListTableWidget(QtWidgets.QTableWidget):
    def __init__(self, columns=3):
        super(ListTableWidget, self).__init__()

        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setColumnCount(columns)
        self.horizontalHeader().setVisible(False)
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setStretchLastSection(True)
        self.setShowGrid(False)


class NoRectDelegate(QtWidgets.QItemDelegate):
    def __init__(self):
        super(NoRectDelegate, self).__init__()

    def drawFocus(self, painter, option, rect):
        option.state &= ~QtWidgets.QStyle.State_HasFocus
        QtWidgets.QItemDelegate.drawFocus(self, painter, option, rect)


def HorizontalSpacerItem(minimum_width=0):
    return QtWidgets.QSpacerItem(
        minimum_width,
        0,
        QtWidgets.QSizePolicy.MinimumExpanding,
        QtWidgets.QSizePolicy.Minimum,
    )


def VerticalSpacerItem():
    return QtWidgets.QSpacerItem(
        0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.MinimumExpanding
    )


class MenuTabWidget(QtWidgets.QWidget):
    """
    A widget that switches between added widgets using a menu on the left.
    This is used to substitute the built-in tab widget, because we need horizontal text buttons on the left.

    Example:
    ```
    menu_tab_widget = MenuTabWidget()
    menu_tab_widget.add_tab("Tab 1", QtWidgets.QWidget())
    menu_tab_widget.add_tab("Tab 2", QtWidgets.QWidget())
    ```
    """

    def __init__(self, *args, **kwargs):
        super(MenuTabWidget, self).__init__(*args, **kwargs)

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._menu_layout = QtWidgets.QVBoxLayout()
        self._menu_layout.setContentsMargins(0, 0, 0, 0)
        self._menu_layout.setSpacing(0)
        self._menu_btn_widget = QtWidgets.QWidget()
        self._menu_btn_widget.setObjectName("MenuTabWidgetMenu")
        self._menu_btn_layout = QtWidgets.QVBoxLayout()
        self._menu_btn_layout.setContentsMargins(0, 0, 0, 0)
        self._menu_btn_layout.setSpacing(0)
        self._menu_btn_widget.setLayout(self._menu_btn_layout)
        self._menu_layout.addWidget(self._menu_btn_widget)
        self._menu_layout.addStretch()

        self._menu_button_group = QtWidgets.QButtonGroup()

        self._layout.addLayout(self._menu_layout)
        self.setLayout(self._layout)

        self.menu_btns = []
        self.tab_widgets = []

        self.set_menu_width()

    def add_tab(self, title: str, widget: QtWidgets.QWidget):
        """
        Add a tab to the tab widget.
        :param title: The title of the tab.
        :param widget: The widget that should be shown when the tab is selected.
        """
        btn = CheckableFlatButton(title)
        btn.clicked.connect(lambda: self.show_tab(widget))
        btn.setFixedHeight(30)
        self.menu_btns.append(btn)
        self._menu_button_group.addButton(btn)
        self._menu_btn_layout.addWidget(btn)
        self.tab_widgets.append(widget)
        self._layout.addWidget(widget)
        widget.hide()

        if len(self.menu_btns) == 1:
            self.select_tab(0)

    def show_tab(self, widget: QtWidgets.QWidget):
        """
        Show the widget. Widg must habe been added with add_tab before.
        """
        for tab_widget in self.tab_widgets:
            tab_widget.hide()
        widget.show()

    def select_tab(self, index):
        """
        Select the tab with the given index.
        """
        self.menu_btns[index].setChecked(True)
        self.show_tab(self.tab_widgets[index])

    def set_menu_width(self, width: int = 100):
        """
        Set the width of the menu on the left.
        """
        self._menu_btn_widget.setFixedWidth(width)



# -*- coding: utf-8 -*-
# Dioptas - GUI program for fast processing of 2D X-ray diffraction data
# Principal author: Clemens Prescher (clemens.prescher@gmail.com)
# Copyright (C) 2014-2019 GSECARS, University of Chicago, USA
# Copyright (C) 2015-2018 Institute for Geology and Mineralogy, University of Cologne, Germany
# Copyright (C) 2019-2020 DESY, Hamburg, Germany
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import time

import numpy as np
from qtpy import QtCore
from colorsys import hsv_to_rgb


class FileNameIterator(QtCore.QObject):
    # TODO create an File Index and then just get the next files according to this.
    # Otherwise searching a network is always to slow...

    def __init__(self, filename=None):
        super(FileNameIterator, self).__init__()
        self.acceptable_file_endings = []
        self.directory_watcher = QtCore.QFileSystemWatcher()
        self.directory_watcher.directoryChanged.connect(self.add_new_files_to_list)
        self.create_timed_file_list = False

        if filename is None:
            self.complete_path = None
            self.directory = None
            self.filename = None
            self.file_list = []
            self.ordered_file_list = []
            self.filename_list = []
        else:
            self.complete_path = os.path.abspath(filename)
            self.directory, self.filename = os.path.split(self.complete_path)
            self.acceptable_file_endings.append(self.filename.split(".")[-1])

    def _get_files_list(self):
        t1 = time.time()
        filename_list = os.listdir(self.directory)
        files = []
        for file in filename_list:
            if self.is_correct_file_type(file):
                files.append(file)
        paths = [os.path.join(self.directory, file) for file in files]
        file_list = [(os.path.getctime(path), path) for path in paths]
        self.filename_list = paths
        print("Time needed  for getting files: {0}s.".format(time.time() - t1))
        return file_list

    def is_correct_file_type(self, filename):
        for ending in self.acceptable_file_endings:
            if filename.endswith(ending):
                return True
        return False

    def _order_file_list(self):
        t1 = time.time()
        self.ordered_file_list = self.file_list
        self.ordered_file_list.sort(key=lambda x: x[0])

        print("Time needed  for ordering files: {0}s.".format(time.time() - t1))

    def update_file_list(self):
        self.file_list = self._get_files_list()
        self._order_file_list()

    def _iterate_file_number(self, path, step, pos=None):
        directory, file_str = os.path.split(path)
        pattern = re.compile(r"\d+")

        match_iterator = pattern.finditer(file_str)

        for ind, match in enumerate(reversed(list(match_iterator))):
            if (pos is None) or (ind == pos):
                number_span = match.span()
                left_ind = number_span[0]
                right_ind = number_span[1]
                number = int(file_str[left_ind:right_ind]) + step
                new_file_str = "{left_str}{number:0{len}}{right_str}".format(
                    left_str=file_str[:left_ind],
                    number=number,
                    len=right_ind - left_ind,
                    right_str=file_str[right_ind:],
                )
                new_file_str_no_leading_zeros = "{left_str}{number}{right_str}".format(
                    left_str=file_str[:left_ind],
                    number=number,
                    right_str=file_str[right_ind:],
                )
                new_complete_path = os.path.join(directory, new_file_str)
                if os.path.exists(new_complete_path):
                    self.complete_path = new_complete_path
                    return new_complete_path
                new_complete_path = os.path.join(
                    directory, new_file_str_no_leading_zeros
                )
                if os.path.exists(new_complete_path):
                    self.complete_path = new_complete_path
                    return new_complete_path
        return None

    def _iterate_folder_number(self, path, step, mec_mode=False):
        directory_str, file_str = os.path.split(path)
        pattern = re.compile(r"\d+")

        match_iterator = pattern.finditer(directory_str)

        for ind, match in enumerate(reversed(list(match_iterator))):
            number_span = match.span()
            left_ind = number_span[0]
            right_ind = number_span[1]
            number = int(directory_str[left_ind:right_ind]) + step
            new_directory_str = "{left_str}{number:0{len}}{right_str}".format(
                left_str=directory_str[:left_ind],
                number=number,
                len=right_ind - left_ind,
                right_str=directory_str[right_ind:],
            )
            print(mec_mode)
            if mec_mode:
                match_file_iterator = pattern.finditer(file_str)
                for ind_file, match_file in enumerate(
                    reversed(list(match_file_iterator))
                ):
                    if ind_file != 2:
                        continue
                    number_span = match_file.span()
                    left_ind = number_span[0]
                    right_ind = number_span[1]
                    number = int(file_str[left_ind:right_ind]) + step
                    new_file_str = "{left_str}{number:0{len}}{right_str}".format(
                        left_str=file_str[:left_ind],
                        number=number,
                        len=right_ind - left_ind,
                        right_str=file_str[right_ind:],
                    )
                new_complete_path = os.path.join(new_directory_str, new_file_str)
                print(new_complete_path)
            else:
                new_complete_path = os.path.join(new_directory_str, file_str)
            if os.path.exists(new_complete_path):
                self.complete_path = new_complete_path
                return new_complete_path

    def get_next_filename(self, step=1, filename=None, mode="number", pos=None):
        if filename is not None:
            self.complete_path = filename

        if self.complete_path is None:
            return None

        if mode == "time":
            time_stat = os.path.getctime(self.complete_path)
            cur_ind = self.ordered_file_list.index((time_stat, self.complete_path))
            # cur_ind = self.ordered_file_list.index(self.complete_path)
            try:
                self.complete_path = self.ordered_file_list[cur_ind + step][1]
                return self.complete_path
            except IndexError:
                return None
        elif mode == "number":
            return self._iterate_file_number(self.complete_path, step, pos)

    def get_previous_filename(self, step=1, filename=None, mode="number", pos=None):
        """
        Tries to get the previous filename.

        :param step:
        :param pos:
        :param mode:
            can have two values either number or mode. Number will decrement the last digits of the file name \
            and time will get the next file by creation time.
        :param filename:
            Filename to get previous number from
        :return:
            either new filename as a string if it exists or None
        """
        if filename is not None:
            self.complete_path = filename

        if self.complete_path is None:
            return None

        if mode == "time":
            time_stat = os.path.getctime(self.complete_path)
            cur_ind = self.ordered_file_list.index((time_stat, self.complete_path))
            # cur_ind = self.ordered_file_list.index(self.complete_path)
            if cur_ind > 0:
                try:
                    self.complete_path = self.ordered_file_list[cur_ind - step][1]
                    return self.complete_path
                except IndexError:
                    return None
        elif mode == "number":
            return self._iterate_file_number(self.complete_path, -step, pos)

    def get_next_folder(self, filename=None, mec_mode=False):
        if filename is not None:
            self.complete_path = filename

        if self.complete_path is None:
            return None
        return self._iterate_folder_number(self.complete_path, 1, mec_mode)

    def get_previous_folder(self, filename=None, mec_mode=False):
        if filename is not None:
            self.complete_path = filename

        if self.complete_path is None:
            return None
        return self._iterate_folder_number(self.complete_path, -1, mec_mode)

    def update_filename(self, new_filename):
        self.complete_path = os.path.abspath(new_filename)
        new_directory, file_str = os.path.split(self.complete_path)
        try:
            self.acceptable_file_endings.append(file_str.split(".")[-1])
        except AttributeError:
            pass
        if self.directory != new_directory:
            if self.directory is not None and self.directory != "":
                self.directory_watcher.removePath(self.directory)
            self.directory_watcher.addPath(new_directory)
            self.directory = new_directory
            if self.create_timed_file_list:
                self.update_file_list()

        if self.create_timed_file_list and self.ordered_file_list == []:
            self.update_file_list()

    def add_new_files_to_list(self):
        """
        checks for new files in folder and adds them to the sorted_file_list
        :return:
        """
        cur_filename_list = os.listdir(self.directory)
        cur_filename_list = [
            os.path.join(self.directory, filename)
            for filename in cur_filename_list
            if self.is_correct_file_type(filename)
        ]
        new_filename_list = [
            filename
            for filename in cur_filename_list
            if filename not in list(self.filename_list)
        ]
        self.filename_list = cur_filename_list
        for filename in new_filename_list:
            creation_time = os.path.getctime(filename)
            if len(self.ordered_file_list) > 0:
                if creation_time > self.ordered_file_list[-1][0]:
                    self.ordered_file_list.append((creation_time, filename))
                else:
                    for ind in range(len(self.ordered_file_list)):
                        if creation_time < self.ordered_file_list[ind][0]:
                            self.ordered_file_list.insert(
                                ind, (creation_time, filename)
                            )
                            break
            else:
                self.ordered_file_list.append((creation_time, filename))


def rotate_matrix_m90(matrix):
    return np.rot90(matrix, -1)


def rotate_matrix_p90(matrix):
    return np.rot90(matrix)


def get_base_name(filename):
    str = os.path.basename(filename)
    if "." in str:
        str = str.split(".")[:-1][0]
    return str


def calculate_color(ind):
    s = 0.8
    v = 0.8
    h = (0.19 * (ind + 2)) % 1
    return np.array(hsv_to_rgb(h, s, v)) * 255


def rgb_to_hex(rgb):
    return "#%02x%02x%02x" % (int(rgb[0]), int(rgb[1]), int(rgb[2]))


def convert_d_to_two_theta(d, wavelength):
    return np.arcsin(wavelength / (2 * d)) / np.pi * 360


def get_partial_index(array, value):
    """
    Calculates the partial index for a value from an array using linear interpolation.
    e.g. with array = [0,1,2,3,4,5] and value = 2.5 it would return 2.5, since it in between the second and third
    element.
    :param array: list or numpy array
    :param value: value for which to get the index
    :return: partial index
    """
    try:
        upper_ind = np.where(array >= value)[0]
        lower_ind = np.where(array < value)[0]
    except TypeError:
        return None

    try:
        spacing = array[upper_ind[0]] - array[lower_ind[-1]]
        new_pos = lower_ind[-1] + (value - array[lower_ind[-1]]) / spacing
    except IndexError:
        return None

    return new_pos


def get_partial_value(array, ind):
    """
    Calculates the value for a non-integer array from an array using linear interpolation.
    e.g. with array = [0,2,4,6,8,10] and value = 2.5 it would return 5, since it is in between the second and third
    element.
    :param array: list or numpy array
    :param ind: float index for which to get value
    """
    if ind < 0 or ind > len(array):
        return None

    step = array[int(np.floor(ind)) + 1] - array[int(np.floor(ind))]
    value = array[int(np.floor(ind))] + (ind - np.floor(ind)) * step
    return value


def reverse_interpolate_two_array(
    value1, array1, value2, array2, delta1=0.1, delta2=0.1
):
    """
    Tries to reverse interpolate two vales from two arrays with the same dimensions, and finds a common index
    for value1 and value2 in their respective arrays. the deltas define the search radius for a close value match
    to the arrays.

    :return: index1, index2
    """
    tth_ind = np.argwhere(np.abs(array1 - value1) < delta1)
    azi_ind = np.argwhere(np.abs(array2 - value2) < delta2)

    tth_ind_ravel = np.ravel_multi_index(
        (tth_ind[:, 0], tth_ind[:, 1]), dims=array1.shape
    )
    azi_ind_ravel = np.ravel_multi_index(
        (azi_ind[:, 0], azi_ind[:, 1]), dims=array2.shape
    )

    common_ind_ravel = np.intersect1d(tth_ind_ravel, azi_ind_ravel)
    result_ind = np.unravel_index(common_ind_ravel, dims=array1.shape)

    while len(result_ind[0]) > 1:
        if np.max(np.diff(array1)) > 0:
            delta1 = np.max(np.diff(array1[result_ind]))

        if np.max(np.diff(array2)) > 0:
            delta2 = np.max(np.diff(array2[result_ind]))

        tth_ind = np.argwhere(np.abs(array1[result_ind] - value1) < delta1)
        azi_ind = np.argwhere(np.abs(array2[result_ind] - value2) < delta2)

        print(result_ind)

        common_ind = np.intersect1d(tth_ind, azi_ind)
        result_ind = (result_ind[0][common_ind], result_ind[1][common_ind])

    return result_ind[0], result_ind[1]
